#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2024, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type
__author__ = ("Trupti A Shetty, Mohamed Rafeek, Madhan Sankaranarayanan, Ajith Andrew J")


DOCUMENTATION = r"""
---
module: rma_workflow_manager
short_description: Manage device replacement workflows in Cisco Catalyst Center.
description:
  - The purpose of this workflow is to provide a streamlined and efficient process for network administrators,
    to initiate Return Material Authorization (RMA) requests for faulty network devices.
    This automation aims to simplify the RMA process, reduce manual effort, and enhance overall operational efficiency.
  - Implement an RMA (Return Material Authorization) workflow within Cisco Catalyst Center,
    enabling a seamless process for returning and replacing faulty network devices.
  - The RMA workflow facilitates the replacement of routers, switches, and Access Points (APs).
  - Allows administrators to mark devices for replacement and track the entire replacement workflow.
  - For routers and switches, the software image, configuration,
    and licenses are restored from the failed device to the replacement device, ensuring minimal disruption.
  - For wireless APs, the replacement device is assigned to the same site, provisioned with the primary wireless controller, RF profile, and AP group settings,
    and placed on the same floor map location in Cisco Catalyst Center as the failed AP.
  - Need to consider the following before doing RMA,
    - Ensure the software image version of the faulty device is imported into the image repository before initiating the replacement process.
    - The faulty device must be in an unreachable state to be eligible for RMA.
    - If the replacement device onboards Cisco Catalyst Center through Plug and Play (PnP), ensure the faulty device is assigned to a user-defined site.
    - The replacement device must not be in a provisioning state during the initiation of the RMA workflow.
    - The AP RMA feature supports only like-to-like replacements,
      meaning the replacement AP must have the same model number and Product ID (PID) as the faulty AP.
    - The replacement AP must have joined the same Cisco Wireless Controller as the faulty AP.
    - Cisco Mobility Express APs acting as wireless controllers are not eligible for replacement through this RMA workflow.
    - Ensure the software image version of the faulty AP is imported into the image repository before initiating the replacement process.
    - The faulty device must be assigned to a user-defined site if the replacement device onboards Cisco Catalyst Center through Plug and Play (PnP).
    - The replacement AP must not be in a provisioning state during the initiation of the RMA workflow.

version_added: '6.6.0'
extends_documentation_fragment:
  - cisco.dnac.workflow_manager_params
author:
  - Trupti A Shetty (@TruptiAShetty)
  - A Mohamed Rafeek (@mohamedrafeek)
  - Madhan Sankaranarayanan (@madhansansel)
  - Ajith Andrew J (@ajithandrewj)

options:
  config_verify:
    description: |
      Set to True to verify the Cisco Catalyst Center configuration after applying the playbook config.
    type: bool
    default: False
  state:
    description: |
      The 'replaced' state is used to indicate the replacement of faulty network devices with
      replacement network device in the workflow.
    type: str
    choices: [ 'replaced' ]
    default: replaced
  ccc_poll_interval:
    description: |
      The interval, in seconds, for polling Cisco Catalyst Center.
    type: int
    default: 2
  resync_retry_count:
    description: |
      The number of times to retry resynchronization.
    type: int
    default: 1000
  resync_retry_interval:
    description: |
      The interval, in seconds, between resynchronization retries.
    type: int
    default: 30
  timeout_interval:
    description: |
      The timeout interval, in seconds, for operations.
    type: int
    default: 100
  config:
    description: |
      A list of faulty and replacement device details for initiating the RMA workflow.
    type: list
    elements: dict
    required: True
    suboptions:
      faulty_device_name:
        description: |
          The name or hostname of the faulty device.
          Example: SJ-EN-9300.cisco.local
        type: str
      faulty_device_ip_address:
        description: |
          The IP address of the faulty device.
          Example: 204.192.3.40
        type: str
      faulty_device_serial_number:
        description: |
          The serial number of the faulty device.
          Example: FJC2327U0S2
        type: str
      replacement_device_ip_address:
        description: |
          The IP address of the replacement device.
          Example: 204.1.2.5
        type: str
      replacement_device_name:
        description: |
          The name or hostname of the replacement device.
          Example: SJ-EN-9300.cisco.local
        type: str
      replacement_device_serial_number:
        description: |
          The serial number of the replacement device.
          Example: FCW2225C020
        type: str

requirements:
  - dnacentersdk >= 2.7.2
  - python >= 3.10

notes:
  - SDK Method used is
    - devices.get_device_detail
    - device_replacement.mark_device_for_replacement
    - device_replacement.deploy_device_replacement_workflow
    - device_replacement.unmark_device_for_replacement
  - Path used is
    - post /dna/intent/api/v1/device-replacement/workflow
    - put  /dna/intent/api/v1/device-replacement/
    - post /dna/intent/api/v1/device-replacement/

  - limitations
  - RMA supports the replacement of similar devices only.
    For instance, a Cisco Catalyst 3650 switch can only be replaced with another Cisco Catalyst 3650 switch.
    The platform IDs of the faulty and replacement devices must match. The model number of a Cisco device can be fetched using the `show version` command.
  - RMA supports the replacement of all switches, routers, and Cisco SD-Access devices, except for the following,
    - Chassis-based Nexus 7700 Series Switches
    - Devices with embedded wireless controllers
    - Cisco Wireless Controllers
  - RMA supports devices with an external SCEP broker PKI certificate.
    The PKI certificate is created and authenticated for the replacement device during the RMA workflow.
    The PKI certificate of the replaced faulty device must be manually deleted from the certificate server.
  - The RMA workflow supports device replacement only if the following conditions are met,
    - Faulty and replacement devices must have the same extension cards.
    - The faulty device must be managed by Catalyst Center with a static IP. (RMA is not supported for devices managed by Catalyst Center with a DHCP IP.)
    - The number of ports in both devices must not vary due to the extension cards.
    - The replacement device must be connected to the same port to which the faulty device was connected.
  - Cisco Catalyst Center does not support legacy license deployment.
  - If the software image installed on the faulty device is earlier than Cisco IOS XE 16.8,
    the same legacy network license must be manually installed on the replacement device.
  - The RMA workflow deregisters the faulty device from Cisco SSM and registers the replacement device with Cisco SSM.
  - Cisco Catalyst Center supports PnP onboarding of the replacement device in a fabric network, except for the following,
    - The faulty device is connected to an uplink device using multiple interfaces.
    - LAN automation using an overlapping pool.
  - If the replacement device onboards through PnP-DHCP functionality,
    ensure the device receives the same IP address after every reload and that the DHCP lease timeout is longer than two hours.
"""

"""
- User can use either one of the below playbook.
"""

EXAMPLES = r"""
- name: RMA workflow for faulty device replacement using device names
  cisco.dnac.rma_workflow_manager:
    dnac_host: "{{ dnac_host }}"
    dnac_username: "{{ dnac_username }}"
    dnac_password: "{{ dnac_password }}"
    dnac_verify: "{{ dnac_verify }}"
    dnac_port: "{{ dnac_port }}"
    dnac_version: "{{ dnac_version }}"
    dnac_debug: "{{ dnac_debug }}"
    dnac_log: true
    dnac_log_level: DEBUG
    config_verify: true
    resync_retry_count: 1000
    resync_retry_interval: 30
    ccc_poll_interval: 2
    timeout_interval: 100
    state: replaced
    config:
      - faulty_device_name: "SJ-EN-9300.cisco.local"
        replacement_device_name: "SJ-EN-9300.cisco-1.local"
  register: result

- name: RMA workflow for faulty device replacement using IP addresses
  cisco.dnac.rma_workflow_manager:
    dnac_host: "{{ dnac_host }}"
    dnac_username: "{{ dnac_username }}"
    dnac_password: "{{ dnac_password }}"
    dnac_verify: "{{ dnac_verify }}"
    dnac_port: "{{ dnac_port }}"
    dnac_version: "{{ dnac_version }}"
    dnac_debug: "{{ dnac_debug }}"
    dnac_log: true
    dnac_log_level: DEBUG
    config_verify: true
    resync_retry_count: 1000
    resync_retry_interval: 30
    ccc_poll_interval: 2
    timeout_interval: 100
    state: replaced
    config:
      - faulty_device_ip_address: "204.192.3.40"
        replacement_device_ip_address: "204.1.2.5"
  register: result

- name: RMA workflow for faulty device replacement using serial numbers
  cisco.dnac.rma_workflow_manager:
    dnac_host: "{{ dnac_host }}"
    dnac_username: "{{ dnac_username }}"
    dnac_password: "{{ dnac_password }}"
    dnac_verify: "{{ dnac_verify }}"
    dnac_port: "{{ dnac_port }}"
    dnac_version: "{{ dnac_version }}"
    dnac_debug: "{{ dnac_debug }}"
    dnac_log: true
    dnac_log_level: DEBUG
    config_verify: true
    resync_retry_count: 1000
    resync_retry_interval: 30
    ccc_poll_interval: 2
    timeout_interval: 100
    state: replaced
    config:
      - faulty_device_serial_number: "FJC2327U0S2"
        replacement_device_serial_number: "FCW2225C020"
  register: result
"""

RETURN = r"""
#Case_1: Marks device for replacement
response_1:
  description: >
    Object with API execution details as returned by the Cisco Catalyst Center Python SDK.
  returned: always
  type: dict
  sample: |
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }

#Case_2: Error while marking device for Replacement.
response_2:
  description: >
    Object with API execution details as returned by the Cisco Catalyst Center Python SDK.
  returned: always
  type: dict
  sample: |
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }

#Case_3: API to trigger RMA workflow that will replace faulty device with replacement device with same configuration and images
response_3:
  description: >
    Object with API execution details as returned by the Cisco Catalyst Center Python SDK.
  returned: always
  type: dict
  sample: |
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }

#Case_4: RMA workflow failed to replace faulty device with replacement device.
response_4:
  description: >
    Object with API execution details as returned by the Cisco Catalyst Center Python SDK.
  returned: always
  type: dict
  sample: |
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }
"""

import re
import json
from ansible_collections.cisco.dnac.plugins.module_utils.dnac import (
    DnacBase,
    validate_list_of_dicts,
    validate_str
)
from ansible.module_utils.basic import AnsibleModule
import time


class DeviceReplacement(DnacBase):
    """Class containing member attributes for rma_workflow_manager module"""

    def __init__(self, module):
        super().__init__(module)
        self.result["response"] = []
        self.supported_states = ["replaced"]
        self.payload = module.params
        self.keymap = {}
        self.faulty_device, self.replacement_device = [], []

    def pprint(self, jsondata):
        return json.dumps(jsondata, indent=4, separators=(',', ': '))

    def validate_input(self):
        """
        Validate the fields provided in the yml files for RMA workflow.
        Checks the configuration provided in the playbook against a predefined specification
        to ensure it adheres to the expected structure and data types based on input.
        Parameters:
            - self (object): An instance of a class used for interacting with the RMA workflow.
        Returns:
            self: An instance of the class with updated attributes:
                - self.msg (str): A message describing the validation result.
                - self.status (str): The status of the validation (either 'success' or 'failed').
                - self.validated_config (list): If successful, a validated version of the 'device_replacements' parameter.
        Description:
            - To use this method, create an instance of the class and call `validate_input` on it.
            - If the validation succeeds, it returns "success" and allows proceeding to the next step.
            - If it fails, `self.status` will be "failed", and `self.msg` will describe the validation issues.
            - The method checks for the presence and validity of various device-related fields such as
            device names, IP addresses, and serial numbers.
            - It removes any None values from the validated configuration.
            - If no configuration is available in the playbook, it returns success with an appropriate message.
        """

        self.log('Validating the Playbook YAML File..', "INFO")

        if not self.config:
            self.status = "success"
            self.msg = "Configuration is not available in the playbook for validation"
            self.log(self.msg, "ERROR")
            return self

        device_list = self.payload.get("config")
        device_list = self.camel_to_snake_case(device_list)

        # Define the expected specification for RMA parameters
        rma_spec = {
            'faulty_device_name': {'required': False, 'type': 'str'},
            'faulty_device_ip_address': {'required': False, 'type': 'str'},
            'replacement_device_name': {'required': False, 'type': 'str'},
            'replacement_device_ip_address': {'required': False, 'type': 'str'},
            'faulty_device_serial_number': {'required': False, 'type': 'str'},
            'replacement_device_serial_number': {'required': False, 'type': 'str'}
        }

        valid_param, invalid_params = validate_list_of_dicts(device_list, rma_spec)
        if invalid_params:
            self.msg = "Invalid parameters in playbook: {0}".format(
                "\n".join(invalid_params)
            )
            self.log(self.msg, "ERROR")
            self.status = "failed"
            return self

        # Remove None values from valid_param
        self.validated_config = []

        for config in valid_param:
            filtered_config = {}
            for key in config:
                if config[key] is not None:
                    filtered_config[key] = config[key]
            self.validated_config.append(filtered_config)

        self.log("Validated config: {0}".format(self.pprint(self.validated_config)), "INFO")
        self.msg = "Successfully validated playbook config params:{0}".format(str(self.validated_config[0]))
        self.log(self.msg, "INFO")
        self.status = "success"
        return self

    def get_want(self, config):
        """
        Get all faulty and replacement device related information from the playbook needed for the RMA workflow
        in Cisco Catalyst Center.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            config (dict): A dictionary containing configuration information for device replacement.
        Returns:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            Retrieves all device replacement configuration details from the playbook config,
            excluding any fields not directly related to the device replacement workflow.
            The extracted information is stored in the 'want' attribute of the instance for
            later use in the workflow. It also performs validation on the configuration parameters.
        """

        want = {}
        want["config"] = {}

        for key in config:
            if config[key] is not None:
                want["config"][key] = config[key]
        self.want = want

        # Perform config validation
        self.validate_device_replacement_params()

        if self.status == "failed":
            self.log("Validation failed. Returning with status 'failed'.", "ERROR")
            return self

        self.log("Desired State (want): {0}".format(str(self.pprint(self.want))), "INFO")
        return self

    def get_have(self):
        """
        Retrieves the current faulty and replacemnet device details from Cisco Catalyst Center.
        Parameters:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            self (object): An instance of the class with updated attributes:
                - self.have: A dictionary containing the current details of the faulty and replacement devices.
                - self.status: The status of the retrieval operation (either 'success' or 'failed').
        Description:
            This method queries Cisco Catalyst Center to check if the specified faulty and
            replacement devices exist. If the devices exist, it retrieves details about them,
            including their IDs and serial numbers. The results are stored in the 'have'
            attribute for later reference in the RMA workflow. If any device is not found
            or an error occurs, it logs the error and updates the status accordingly.
        """

        # Check if 'want' dictionary is valid
        if not self.want or not self.want.get("config"):
            self.msg = "Invalid or missing 'want' dictionary"
            self.log(self.msg, "ERROR")
            self.status = "failed"
            return self

        have = {}
        config = self.want["config"]
        identifier_keys = [
            ("faulty_device_serial_number", "replacement_device_serial_number"),
            ("faulty_device_serial_number", "replacement_device_name"),
            ("faulty_device_serial_number", "replacement_device_ip_address"),
            ("faulty_device_name", "replacement_device_serial_number"),
            ("faulty_device_name", "replacement_device_name"),
            ("faulty_device_name", "replacement_device_ip_address"),
            ("faulty_device_ip_address", "replacement_device_ip_address"),
            ("faulty_device_ip_address", "replacement_device_name"),
            ("faulty_device_ip_address", "replacement_device_serial_number")
        ]

        valid_identifier_found = False

        # Iterate through identifier keys to find valid device combinations
        for faulty_key, replacement_key in identifier_keys:
            faulty_identifier = config.get(faulty_key)
            replacement_identifier = config.get(replacement_key)

            if faulty_identifier and replacement_identifier:
                valid_identifier_found = True

                # Check if faulty device exists
                faulty_device = self.device_exists(faulty_identifier, faulty_key)

                if not faulty_device:
                    self.msg = "Faulty device '{0}' not found in Cisco Catalyst Center".format(faulty_identifier)
                    self.log(self.msg, "ERROR")
                    self.status = "failed"
                    return self

                have["faulty_device_id"] = faulty_device.get("device_id")
                have["faulty_device_serial_number"] = faulty_device.get("serial_number")
                have["faulty_device_name"] = faulty_device.get("device_name")
                have["faulty_device_reachability_status"] = faulty_device.get("reachability_status")
                have["faulty_device_platform_id"] = faulty_device.get("platform_id")
                have[faulty_key] = faulty_identifier
                have["faulty_device_exists"] = True
                self.log("Faulty device '{0}' found in Cisco Catalyst Center".format(faulty_identifier), "INFO")

                # Check if replacement device exists
                replacement_device = self.device_exists(replacement_identifier, replacement_key)

                if not replacement_device:
                    self.log("Replacement device '{0}' not found in inventory, checking in PnP...", "DEBUG")
                    replacement_device = self.pnp_device_exists(replacement_identifier, replacement_key)

                    if not replacement_device:
                        self.msg = "Replacement device '{0}' not found in PnP".format(replacement_identifier)
                        self.log(self.msg, "ERROR")
                        self.status = "failed"
                        return self

                have["replacement_device_id"] = replacement_device.get("device_id")
                have["replacement_device_serial_number"] = replacement_device.get("serial_number")
                have["replacement_device_name"] = replacement_device.get("device_name")
                have["replacement_device_reachability_status"] = replacement_device.get("reachability_status")
                have["replacement_device_platform_id"] = replacement_device.get("platform_id")
                have["is_pnp_replacement_device"] = replacement_device.get("is_pnp_device")
                have[replacement_key] = replacement_identifier
                have["replacement_device_exists"] = True
                self.log("Replacement device '{0}' found in Cisco Catalyst Center".format(replacement_identifier), "INFO")
                break

        # Check if any valid identifier combination was not found
        if not valid_identifier_found:
            provided_identifiers = {
                key: value
                for key, value in config.items()
                if key in [item for sublist in identifier_keys for item in sublist] and value
            }
            self.msg = "No valid device combination found in config. Provided values in config: {0}".format(provided_identifiers)
            self.log(self.msg, "ERROR")
            self.status = "failed"
            return self

        self.have = have

        if not self.have:
            self.msg = "No valid device information found in config"
            self.log(self.msg, "ERROR")
            self.status = "failed"
        else:
            self.msg = "Successfully retrieved device details: {0}".format(self.pprint(config))
            self.log("Current State (have): {0}".format(self.pprint(self.have)), "INFO")
            self.log(self.msg, "INFO")
            self.status = "success"

        return self

    def rma_device_replacement_pre_check(self):
        """
        Performs a pre-check for RMA device replacement to ensure compatibility and reachability.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This method verifies that the faulty device and the replacement device belong to the same family and series,
            ensuring they are compatible for replacement. It also checks the network reachability of the replacement device.
            If both checks pass, the method logs a success message and proceeds. If either check fails, it logs an error,
            updates the status to 'failed', and returns the instance for further handling in the RMA workflow.
        """

        if self.have["faulty_device_platform_id"] != self.have["replacement_device_platform_id"]:
            self.msg = (
                "The faulty device and the replacement device do not belong to the same platform, family and series."
                " These attributes must match for a valid replacement."
            )
            self.log(self.msg, "ERROR")
            self.status = "failed"
            return self

        self.log("The faulty device and the replacement device belong to the same platform, family and series.", "DEBUG")

        if not self.have["is_pnp_replacement_device"]:
            if self.have["replacement_device_reachability_status"] != "Reachable":
                self.msg = "The replacement device is not reachable. Unable to proceed with the RMA device replacement."
                self.log(self.msg, "ERROR")
                self.status = "failed"
                return self

            self.log("The replacement device '{0}' is reachable.".format(self.have.get("replacement_device_name")), "DEBUG")

        return self

    def pnp_device_exists(self, identifier, identifier_type):
        """
        Check if a pnp device exists in Cisco Catalyst Center and return its device ID, device_name, serial_number and platform_id.
        Parameters:
            - self (object): An instance of the class containing the method.
            - identifier (str): The identifier of the device to check.
            - identifier_type (str): The type of identifier (name, ip_address, or serial_number).
        Returns:
            - dict: A dict containing the device ID, device_name, serial_number and platform_id if the device
              is found or empty dict if device not found.
        Description:
            This method queries Cisco Catalyst Center to check if a specified device exists based on the provided identifier.
            It constructs the appropriate query parameters based on the identifier type (hostname, IP address, or serial number).
            The method then sends a request to Cisco Catalyst Center using the 'get_device_list' function.
            If the device is found and both ID and serial number are available, it returns these as a tuple.
            If the device is not found, lacks necessary information, or if an error occurs during the process,
            it logs an appropriate error message and returns empty dict.
            This method is used to verify the existence of both faulty and replacement devices in the RMA workflow.
        """
        params = {}

        if identifier_type.endswith("_name"):
            params["hostname"] = identifier
        elif identifier_type.endswith("_serial_number"):
            params["serialNumber"] = identifier
        else:
            self.log("Invalid identifier type provided", "ERROR")
            return {}

        try:
            response = self.dnac._exec(
                family="device_onboarding_pnp",
                function='get_device_list',
                op_modifies=False,
                params=params
            )
            self.log("Received API response from 'get_device_list': {0}".format(self.pprint(response)), "DEBUG")

            if response:
                device = response[0]
                device_info = device.get('deviceInfo', {})
                device_param_list = {
                    "device_id": device.get('id'),
                    "serial_number": device_info.get('serialNumber'),
                    "device_name": device_info.get('hostname'),
                    "platform_id": device_info.get('pid'),
                    "is_pnp_device": True
                }

                if device_param_list:
                    return device_param_list
                self.log("Device found but ID or serial number missing", "ERROR")
            else:
                self.log("Device not found in Cisco Catalyst Center", "ERROR")
        except Exception as e:
            self.log("Exception occurred while querying device: {0}".format(str(e)), "ERROR")

        return {}

    def device_exists(self, identifier, identifier_type):
        """
        Check if a device exists in Cisco Catalyst Center and return its device ID, serial_number, device_name, reachability_status, platform_id.
        Parameters:
            - self (object): An instance of the class containing the method.
            - identifier (str): The identifier of the device to check.
            - identifier_type (str): The type of identifier (name, ip_address, or serial_number).
        Returns:
            - dict: A dict containing the device ID, serial_number, device_name, reachability_status, platform_id if the device
              is found or empty dict if device not found.
        Description:
            This method queries Cisco Catalyst Center to check if a specified device exists based on the provided identifier.
            It constructs the appropriate query parameters based on the identifier type (hostname, IP address, or serial number).
            The method then sends a request to Cisco Catalyst Center using the 'get_device_list' function.
            If the device is found and both ID and serial number are available, it returns these as a tuple.
            If the device is not found, lacks necessary information, or if an error occurs during the process,
            it logs an appropriate error message and returns empty dict.
            This method is used to verify the existence of both faulty and replacement devices in the RMA workflow.
        """
        params = {}
        if identifier_type.endswith("_name"):
            params["hostname"] = identifier
        elif identifier_type.endswith("_ip_address"):
            params["managementIpAddress"] = identifier
        elif identifier_type.endswith("_serial_number"):
            params["serialNumber"] = identifier
        else:
            self.log("Invalid identifier type provided", "ERROR")
            return {}

        try:
            response = self.dnac._exec(
                family="devices",
                function='get_device_list',
                op_modifies=False,
                params=params
            )
            self.log("Received API response from 'get_device_list': {0}".format(self.pprint(response)), "DEBUG")
            device_param_list = {}

            if response and response.get('response'):
                if len(response['response']) > 0:
                    device = response['response'][0]
                    device_param_list["device_id"] = device.get('id')
                    device_param_list["serial_number"] = device.get('serialNumber')
                    device_param_list["device_name"] = device.get('hostname')
                    device_param_list["reachability_status"] = device.get('reachabilityStatus')
                    device_param_list["platform_id"] = device.get('platformId')
                    device_param_list["is_pnp_device"] = False

                    if device_param_list:
                        return device_param_list
                    self.log("Device found but ID or serial number missing", "ERROR")
                else:
                    self.log("Device not found in Cisco Catalyst Center", "ERROR")
            else:
                self.log("No valid response received from Cisco Catalyst Center", "ERROR")
        except Exception as e:
            self.log("Exception occurred while querying device: {0}".format(str(e)), "ERROR")

        return {}

    def validate_device_replacement_params(self):
        """
        Addtional validation for the faulty and replacemnet device parameters.
        Parameters:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            - config (dict): A dictionary containing the faulty and replacement device details.
        Returns:
           The method returns an instance of the class with updated attributes:
                - self.msg: A message describing the validation result.
                - self.status: The status of the validation (either 'success' or 'failed').
        Description:
            This method validates the configuration parameters for the faulty and replacement devices.
            It checks the device names, IP addresses, and serial numbers for correctness. If validation
            fails, it updates 'self.status' to 'failed' and logs the issues in 'self.msg'. If validation
            succeeds, it sets 'self.status' to 'success'.
        """

        errormsg = []
        config = self.want.get("config", {})

        # Validate device names
        for name_field in ['faulty_device_name', 'replacement_device_name']:
            if config.get(name_field):
                param_spec = dict(type="str", length_max=255)
                validate_str(config[name_field], param_spec, name_field, errormsg)

        # Validate IP addresses
        for ip_field in ['faulty_device_ip_address', 'replacement_device_ip_address']:
            if config.get(ip_field):
                if not self.is_valid_ipv4(config[ip_field]):
                    errormsg.append("{0}: Invalid IP Address '{1}' in playbook".format(ip_field, config[ip_field]))

        # Validate serial numbers
        serial_regex = re.compile(r'^[A-Z0-9]{11}$')
        for serial_field in ['faulty_device_serial_number', 'replacement_device_serial_number']:
            if config.get(serial_field):
                if not serial_regex.match(config[serial_field]):
                    errormsg.append("{0}: Invalid Serial Number '{1}' in playbook.".format(
                        serial_field, config[serial_field]))

        if errormsg:
            self.msg = "Invalid parameters in playbook config: '{0}' ".format(str("\n".join(errormsg)))
            self.log(self.msg, "ERROR")
            self.status = "failed"
            return self

        self.msg = "Successfully validated config params:{0}".format(self.pprint(config))
        self.log(self.msg, "INFO")
        self.status = "success"
        return self

    def device_ready_for_replacement_check(self):
        """
        Checks if the faulty device is ready for replacement.
        Parameters:
            - self (object): An instance of the class that interacts with Cisco Catalyst Center and contains device details.
        Returns:
            bool:
                - True if the faulty device is found and is in the "READY-FOR-REPLACEMENT" state.
                - False if the faulty device is not found or is not in the "READY-FOR-REPLACEMENT" state.
        Description:
            This method retrieves a list of devices marked for replacement from Cisco Catalyst Center
            using the `device_replacement` API. It iterates through the returned devices to find
            the specified faulty device based on its serial number, which is stored in the `self.have` attribute.
            If the faulty device is found and its status is "READY-FOR-REPLACEMENT", the method logs a debug message
            indicating that the device is already marked for replacement and returns `True`.
            If the device is not in the "READY-FOR-REPLACEMENT" state or is not found, it returns `False`.
        """
        response = self.dnac._exec(
            family="device_replacement",
            function='return_replacement_devices_with_details'
        )
        devices = response.get("response", [])
        self.log("Received API response from 'return_replacement_devices_with_details': {0}".format(self.pprint(response)), "DEBUG")

        for device in devices:
            if device.get("faultyDeviceSerialNumber") == self.have.get("faulty_device_serial_number"):
                if device.get("replacementStatus") == "READY-FOR-REPLACEMENT":
                    self.have["device_replacement_id"] = device.get("id")
                    return True

        return False

    def mark_faulty_device_for_replacement(self):
        """
        Mark the faulty device for replacement in Cisco Catalyst Center.
        Parameters:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This method marks a faulty device for replacement in Cisco Catalyst Center. It performs the following steps:
            - Checks if the faulty device ID is available.
            - Prepares the payload for the API call.
            - Sends a request to Cisco Catalyst Center to mark the device for replacement.
            - Processes the API response and extracts the task ID.
            - Uses the check_rma_task_status method to monitor the task status.
            - Updates the status, msg, and result attributes based on the task result.
            - Handles any exceptions that occur during the process.
        """
        is_ready_for_replacement = self.device_ready_for_replacement_check()
        if not is_ready_for_replacement:
            import_params = dict(
                payload=[{
                    "faultyDeviceId": self.have.get("faulty_device_id"),
                    "replacementStatus": "MARKED-FOR-REPLACEMENT"
                }],
            )

            try:
                response = self.dnac._exec(
                    family="device_replacement",
                    function='mark_device_for_replacement',
                    params=import_params
                )
                self.log("Received API response from 'mark_device_for_replacement': {0}".format(str(response)), "DEBUG")
                task_id = response.get("response", {}).get("taskId")
                task_result = self.check_rma_task_status(
                    task_id,
                    "Device marked for replacement successfully",
                    "Error while marking device for replacement"
                )
                self.status = task_result["status"]
                self.msg = task_result["msg"]
                if self.status == "success":
                    self.result['changed'] = True
                self.device_ready_for_replacement_check()
                return self

            except Exception as e:
                self.status = "failed"
                self.msg = "Exception occurred while marking device for replacement: {0}".format(str(e))
                self.log(self.msg, "ERROR")

        self.log("The device '{0}' is already in the 'READY-FOR-REPLACEMENT' state.".format(self.have.get("faulty_device_name")), "DEBUG")
        return self

    def get_diff_replaced(self, config):
        """
        Replace a faulty device with a new device in Cisco Catalyst Center.
        Parameters:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            - config (dict): Configuration dictionary (not used in this method, but included for consistency).
        Returns:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This method replaces a faulty device with a new device in Cisco Catalyst Center. It performs the following steps:
            - Checks if both faulty and replacement device serial numbers are available.
            - Prepares the payload for the API call.
            - Sends a request to Cisco Catalyst Center to deploy the device replacement workflow.
            - Processes the API response and extracts the task ID.
            - Uses the check_replacement_status method to monitor the replacement status.
            - Uses the check_rma_task_status method to monitor the task status.
            - Updates the status, msg, and result attributes based on the task result.
            - If the replacement fails, it attempts to unmark the faulty device.
            - Handles any exceptions that occur during the process.
        """

        import_params = dict(
            payload={
                "faultyDeviceSerialNumber": self.have.get("faulty_device_serial_number"),
                "replacementDeviceSerialNumber": self.have.get("replacement_device_serial_number")
            }
        )

        self.log("Replacing device with parameters: {0}".format(self.pprint(import_params)), "INFO")

        try:
            response = self.dnac._exec(
                family="device_replacement",
                function='deploy_device_replacement_workflow',
                op_modifies=True,
                params=import_params
            )
            self.log("Received API response from 'deploy_device_replacement_workflow': {0}".format(self.pprint(response)), "DEBUG")
            task_id = response.get("response", {}).get("taskId")

            # Monitor the task status using check_rma_task_status
            task_result = self.check_rma_task_status(
                task_id,
                "Device replacement task initiated successfully",
                "Error in device replacement task initiation"
            )
            if task_result["status"] != "success":
                self.status = "failed"
                error_msg = "Device replacement task failed: {0}".format(task_result["msg"])
                self.log(error_msg, "ERROR")
                self.result['msg'] = error_msg
                # Attempt to unmark the device
                self.log("Attempting to unmark the device after failure", "INFO")
                unmark_result = self.unmark_device_for_replacement()
                # Combine both error messages
                self.msg = "{0} | Unmarking result: {1}".format(error_msg, unmark_result.msg)
                self.log(self.msg, "ERROR")
                self.result['msg'] = self.msg
                return self

            # If task is initiated successfully, monitor the replacement status
            self.task_id = task_id
            replacement_result = self.monitor_replacement_status()
            self.status = replacement_result["status"]
            self.msg = replacement_result["msg"]
            if self.status != "success":
                self.status = "failed"
                self.result['msg'] = self.msg
                # Attempt to unmark the device
                self.log("Attempting to unmark the device after failure", "INFO")
                unmark_result = self.unmark_device_for_replacement()
                self.msg = "{0} | Unmarking result: {1}".format(self.msg, unmark_result.msg)
                self.log(self.msg, "ERROR")
                self.result['msg'] = self.msg
                return self

            self.faulty_device.append(self.have.get("faulty_device_name"))
            self.replacement_device.append(self.have.get("replacement_device_name"))
            self.result['changed'] = True
            self.result['msg'] = self.msg

        except Exception as e:
            self.status = "failed"
            error_msg = "Exception occurred during device replacement "
            self.log(error_msg, "ERROR")
            # Attempt to unmark the device
            self.log("Attempting to unmark the device after exception", "INFO")
            unmark_result = self.unmark_device_for_replacement()
            # Combine both error messages
            self.msg = "{0} | Unmarking result: {1}".format(error_msg, unmark_result.msg)
            self.log(self.msg, "ERROR")
            self.result['msg'] = self.msg
            self.result['response'] = []

        return self

    def monitor_replacement_status(self):
        """
        Monitor the status of the device replacement task in Cisco Catalyst Center.
        Parameters:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            - dict: A dictionary containing the status and message of the replacement task.
        Description:
            This method monitors the status of a device replacement task in Cisco Catalyst Center. It performs the following steps:
            - Initializes retry count and interval for checking task status.
            - Enters a loop to periodically check the task status:
                - Retrieves task details using the get_task_details method.
                - Checks if the task has completed successfully:
                    - If successful, updates result attributes and returns success status.
                - Checks if the task has encountered an error:
                    - If error occurred, updates result attributes and returns failed status.
                - If task is still in progress:
                    - Logs the progress and waits for the retry interval before checking again.
            - If the maximum number of retries is reached without a definitive result:
                - Sets the status to failed and logs a timeout message.
            - Handles various scenarios of task completion, failure, or timeout.
            - Returns a dictionary with the final status and message of the replacement task.
        """

        resync_retry_count = self.params.get('resync_retry_count')
        resync_retry_interval = self.params.get('resync_retry_interval')
        while resync_retry_count:
            task_details = self.get_task_details(self.task_id)
            self.log("Task Details: {0}".format(self.pprint(task_details)), "DEBUG")

            if task_details.get("endTime") is not None:
                if task_details.get("isError") is False:
                    self.result['changed'] = True
                    self.msg = "Device replacement completed successfully: {0}".format(task_details.get("progress"))
                    self.log(self.msg, "INFO")
                    self.result['task_response'] = {
                        "replacement_task_response": task_details,
                        "replacement_status": self.msg
                    }
                    return {"status": "success", "msg": self.msg}

                self.result['changed'] = False
                self.status = "failed"
                self.msg = "Error in device replacement: {0}".format(task_details.get("progress"))
                self.log(self.msg, "ERROR")
                self.result['task_response'] = {
                    "replacement_task_response": task_details,
                    "replacement_status": self.msg
                }
                return {"status": "failed", "msg": self.msg}

            self.log("RMA workflow in progress: {0}".format(task_details.get("progress")), "INFO")
            time.sleep(resync_retry_interval)
            resync_retry_count -= 1

        # If we've exhausted all retries without a definitive result
        self.status = "failed"
        self.msg = "Device replacement monitoring timed out after {0} attempts".format(self.params.get('dnac_api_task_timeout'))
        self.log(self.msg, "ERROR")
        return {"status": "failed", "msg": self.msg}

    def unmark_device_for_replacement(self):
        """
        Unmark the faulty device for replacement in Cisco Catalyst Center only when replacing of faulty device to replacement device fails.
        Parameters:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This method unmarks a faulty device for replacement in Cisco Catalyst Center. It performs the following steps:
            - Checks if the faulty device ID is available.
            - Prepares the payload for the API call.
            - Sends a request to Cisco Catalyst Center to unmark the device for replacement.
            - Processes the API response and extracts the task ID.
            - Uses the check_rma_task_status method to monitor the task status.
            - Updates the status, msg, and result attributes based on the task result.
            - Handles any exceptions that occur during the process.
        """
        self.log("Unmarking device for replacement...")
        device_id = self.get_ready_for_replacement_device_id()

        import_params = dict(
            payload=[{
                "id": device_id,
                "replacementStatus": "NON-FAULTY"
            }],
        )

        try:
            response = self.dnac._exec(
                family="device_replacement",
                function='unmark_device_for_replacement',
                op_modifies=True,
                params=import_params
            )
            self.log("Received API response from 'unmark_device_for_replacement': {0}".format(self.pprint(response)), "DEBUG")
            task_id = response.get("response", {}).get("taskId")
            task_result = self.check_rma_task_status(
                task_id,
                "Device unmarked for replacement successfully",
                "Error while unmarking device for replacement"
            )
            self.status = task_result["status"]
            self.msg = "RMA failed to replace the device: {0}".format(task_result["msg"])

        except Exception:
            self.status = "failed"
            self.msg = "RMA failed to replace the device: No device found for unmarking replacement"
            self.log(self.msg, "ERROR")
        return self

    def get_ready_for_replacement_device_id(self):
        """
        Retrieves the ID of the first device marked as "READY-FOR-REPLACEMENT" in Cisco Catalyst Center.
        Parameters:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            - device_id (str or None): The ID of the first device ready for replacement, or None if no such device is found.
        Description:
            - This method fetches a list of devices with their replacement status from Cisco Catalyst Center.
            - It then checks for the first device with a "READY-FOR-REPLACEMENT" status and returns its ID.
            - The method exits early if such a device is found.
        """
        response = self.dnac._exec(
            family="device_replacement",
            function='return_replacement_devices_with_details'
        )
        devices = response.get("response", [])
        for device in devices:
            if device.get("replacementStatus") == "READY-FOR-REPLACEMENT":
                device_id = device.get("id")
                self.log("Found ready-for-replacement device with ID: {0}".format(device_id))
                return device_id

        self.log("No devices found with status 'READY-FOR-REPLACEMENT'.")
        return None

    def check_rma_task_status(self, task_id, success_message, error_prefix):
        """
        Check the status of an RMA task in Cisco Catalyst Center.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            task_id (str): The ID of the task to monitor.
            success_message (str): The message to log on successful completion.
            error_prefix (str): The prefix for the error message if the task fails.
        Returns:
            dict: A dictionary containing the status and message of the task result.
        Description:
            This method checks the status of an RMA task in Cisco Catalyst Center. It performs the following steps:
            - Continuously polls the task status using the get_task_details method.
            - Checks if the task has completed successfully or encountered an error.
            - Logs appropriate messages based on the task outcome.
            - Returns a dictionary with the task status and message.
            - Implements a delay between status checks to avoid overwhelming the API.
        """

        ccc_poll_interval = self.params.get('ccc_poll_interval')
        timeout_interval = self.params.get('timeout_interval')
        while timeout_interval > 0:
            task_details = self.get_task_details(task_id)
            self.log(task_details)
            if task_details.get("isError"):
                error_message = task_details.get("failureReason", "{0}: Task failed.".format(error_prefix))
                self.log(error_message, "ERROR")
                return {"status": "failed", "msg": error_message}

            if 'progress' in task_details:
                progress = task_details['progress'].lower()

                if 'successful' in progress:
                    self.log(success_message, "INFO")
                    return {"status": "success", "msg": progress}

            time.sleep(ccc_poll_interval)
            timeout_interval -= ccc_poll_interval

    def update_rma_profile_messages(self):
        """
        Updates and logs messages based on the status of RMA device replacements.
        Args:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            self (object): The current instance of the class with updated `result` and `msg` attributes.
        Description:
            This method generates and updates status messages regarding the RMA (Return Material Authorization) device replacement process.
            It checks if there are any faulty and replacement devices specified for replacement. If both are present, it constructs a
            success message detailing the completion of the replacement process for the faulty device(s) with the corresponding replacement device(s).
            If no faulty or replacement devices are found, it sets a message indicating that no replacements were performed.
            The method then updates the `result` attribute with the status of the operation (`changed` set to True if replacements occurred)
            and logs the final message using the appropriate log level. The constructed message is also stored in `result["response"]`
            for further reference.
        """
        self.result["changed"] = False
        result_msg_list = []

        if self.faulty_device and self.replacement_device:
            device_replacement_msg = (
                "Device replacement was successfully completed for the faulty device(s) '{0}',"
                " with the replacement device(s) '{1}'.".format("', '".join(self.faulty_device), "', '".join(self.replacement_device))
            )
            result_msg_list.append(device_replacement_msg)

        if result_msg_list:
            self.result["changed"] = True
            self.msg = " ".join(result_msg_list)
        else:
            self.msg = "No changes were made. No RMA device replacement were performed in Cisco Catalyst Center."

        self.log(self.msg, "INFO")
        self.result["response"] = self.msg

        return self

    def verify_diff_replaced(self, config):
        """
        Verify the device replacement status in Cisco Catalyst Center.
        Parameters:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This method verifies the replacement status of a device in Cisco Catalyst Center. It performs the following steps:
            - Prepares the payload for the API call using the replacement device serial number.
            - Checks if the replacement device serial number is available.
            - Sends a request to Cisco Catalyst Center to get details of replacement devices.
            - Processes the API response to find the matching device.
            - Logs the replacement status of the matching device, if found.
            - Handles any exceptions that occur during the process.
            - Always returns self to maintain method chaining.
        """

        replacement_device_serial = self.have.get("replacement_device_serial_number")
        if not replacement_device_serial:
            self.log("Replacement device serial number is missing", "WARNING")
            return self

        import_params = {"replacementDeviceSerialNumber": replacement_device_serial}

        try:
            response = self.dnac._exec(
                family="device_replacement",
                function='return_replacement_devices_with_details',
                params=import_params
            )
            devices = response.get("response", [])
            replacement_status = None
            for device in devices:
                if device.get("id") == self.have.get("device_replacement_id"):
                    replacement_status = device
            self.log("Replacement status: {0}".format(self.pprint(replacement_status)), "INFO")
        except Exception as e:
            self.log("Error getting replacement status: {0}".format(str(e)), "ERROR")

        return self


def main():
    """ main entry point for module execution
    """
    # Basic Ansible type check and assigning defaults.
    device_replacement_spec = {
        'dnac_host': {'required': True, 'type': 'str'},
        'dnac_port': {'type': 'str', 'default': '443'},
        'dnac_username': {'type': 'str', 'default': 'admin', 'aliases': ['user']},
        'dnac_password': {'type': 'str', 'no_log': True},
        'dnac_verify': {'type': 'bool', 'default': 'True'},
        'dnac_version': {'type': 'str', 'default': '2.2.3.3'},
        'dnac_debug': {'type': 'bool', 'default': False},
        'dnac_log': {'type': 'bool', 'default': False},
        'dnac_log_level': {'type': 'str', 'default': 'WARNING'},
        "dnac_log_file_path": {"type": 'str', "default": 'dnac.log'},
        'config_verify': {'type': 'bool', "default": False},
        "dnac_log_append": {"type": 'bool', "default": True},
        'dnac_api_task_timeout': {'type': 'int', "default": 1200},
        'dnac_task_poll_interval': {'type': 'int', "default": 2},
        'resync_retry_count': {'type': 'int', 'default': 1000},
        'resync_retry_interval': {'type': 'int', 'default': 30},
        'ccc_poll_interval': {'type': 'int', 'default': 2},
        'timeout_interval': {'type': 'int', 'default': 100},
        'config': {'required': True, 'type': 'list', 'elements': 'dict'},
        'validate_response_schema': {'type': 'bool', 'default': True},
        'state': {'default': 'replaced', 'choices': ['replaced']}
    }
    module = AnsibleModule(
        argument_spec=device_replacement_spec,
        supports_check_mode=True
    )

    ccc_device_replacement = DeviceReplacement(module)
    state = ccc_device_replacement.params.get("state")

    ccc_device_replacement.validate_input().check_return_status()
    config_verify = ccc_device_replacement.params.get("config_verify")

    for config in ccc_device_replacement.validated_config:
        ccc_device_replacement.reset_values()
        ccc_device_replacement.get_want(config).check_return_status()
        ccc_device_replacement.get_have().check_return_status()
        ccc_device_replacement.rma_device_replacement_pre_check().check_return_status()
        ccc_device_replacement.mark_faulty_device_for_replacement().check_return_status()
        ccc_device_replacement.get_diff_state_apply[state](config).check_return_status()
        if config_verify:
            ccc_device_replacement.verify_diff_state_apply[state](config).check_return_status()

    ccc_device_replacement.update_rma_profile_messages().check_return_status()

    module.exit_json(**ccc_device_replacement.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = r"""
---
module: system_issue_definitions_count_info
short_description: Information module for System Issue Definitions Count Info
description:
- This module represents an alias of the module system_issue_definitions_count_v1_info
version_added: '6.15.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  deviceType:
    description:
    - >
      DeviceType query parameter. These are the device families/types supported for system issue definitions. If
      no input is made on device type, all device types are considered.
    type: str
  profileId:
    description:
    - >
      ProfileId query parameter. The profile identier to fetch the profile associated issue defintions. The
      default is `global`. Please refer Network design profiles documentation for more details.
    type: str
  id:
    description:
    - >
      Id query parameter. The definition identifier. Examples id=015d9cba-4f53-4087-8317-7e49e5ffef46 (single
      entity id request) id=015d9cba-4f53-4087-8317-7e49e5ffef46&id=015d9cba-4f53-4087-8317-7e49e5ffef47 (multiple
      ids in the query param).
    type: str
  name:
    description:
    - >
      Name query parameter. The list of system defined issue names. (Ex."BGP_Down") Examples name=BGP_Down (single
      entity uuid requested) name=BGP_Down&name=BGP_Flap (multiple issue names separated by & operator).
    type: str
  priority:
    description:
    - >
      Priority query parameter. Issue priority, possible values are P1, P2, P3, P4. `P1` A critical issue that
      needs immediate attention and can have a wide impact on network operations. `P2` A major issue that can
      potentially impact multiple devices or clients. `P3` A minor issue that has a localized or minimal impact.
      `P4` A warning issue that may not be an immediate problem but addressing it can optimize the network
      performance.
    type: str
  issueEnabled:
    description:
    - IssueEnabled query parameter. The enablement status of the issue definition, either true or false.
    type: bool
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Issues GetTheCountOfSystemDefinedIssueDefinitionsBasedOnProvidedFiltersV1
  description: Complete reference of the GetTheCountOfSystemDefinedIssueDefinitionsBasedOnProvidedFiltersV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!get-the-count-of-system-defined-issue-definitions-based-on-provided-filters
notes:
  - SDK Method used are
    issues.Issues.get_the_count_of_system_defined_issue_definitions_based_on_provided_filters_v1,

  - Paths used are
    get /dna/intent/api/v1/systemIssueDefinitions/count,
  - It should be noted that this module is an alias of system_issue_definitions_count_v1_info

"""

EXAMPLES = r"""
- name: Get all System Issue Definitions Count Info
  cisco.dnac.system_issue_definitions_count_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    deviceType: string
    profileId: string
    id: string
    name: string
    priority: string
    issueEnabled: True
  register: result

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "count": 0
      },
      "version": "string"
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: sda_port_assignments_v1
short_description: Resource module for Sda Port Assignments V1
description:
- Manage operations create, update and delete of the resource Sda Port Assignments V1.
- Adds port assignments based on user input.
- Deletes a port assignment based on id.
- Deletes port assignments based on user input.
- Updates port assignments based on user input.
version_added: '6.14.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  dataVlanName:
    description: DataVlanName query parameter. Data VLAN name of the port assignment.
    type: str
  fabricId:
    description: FabricId query parameter. ID of the fabric the device is assigned to.
    type: str
  id:
    description: Id path parameter. ID of the port assignment.
    type: str
  interfaceName:
    description: InterfaceName query parameter. Interface name of the port assignment.
    type: str
  networkDeviceId:
    description: NetworkDeviceId query parameter. Network device ID of the port assignment.
    type: str
  payload:
    description: Sda Port Assignments's payload.
    elements: dict
    suboptions:
      authenticateTemplateName:
        description: Authenticate template name of the port assignment.
        type: str
      connectedDeviceType:
        description: Connected device type of the port assignment.
        type: str
      dataVlanName:
        description: Data VLAN name of the port assignment.
        type: str
      fabricId:
        description: ID of the fabric the device is assigned to.
        type: str
      interfaceDescription:
        description: Interface description of the port assignment.
        type: str
      interfaceName:
        description: Interface name of the port assignment.
        type: str
      networkDeviceId:
        description: Network device ID of the port assignment.
        type: str
      securityGroupName:
        description: Security group name of the port assignment.
        type: str
      voiceVlanName:
        description: Voice VLAN name of the port assignment.
        type: str
    type: list
  voiceVlanName:
    description: VoiceVlanName query parameter. Voice VLAN name of the port assignment.
    type: str
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for SDA AddPortAssignmentsV1
  description: Complete reference of the AddPortAssignmentsV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!add-port-assignments
- name: Cisco DNA Center documentation for SDA DeletePortAssignmentByIdV1
  description: Complete reference of the DeletePortAssignmentByIdV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!delete-port-assignment-by-id
- name: Cisco DNA Center documentation for SDA DeletePortAssignmentsV1
  description: Complete reference of the DeletePortAssignmentsV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!delete-port-assignments
- name: Cisco DNA Center documentation for SDA UpdatePortAssignmentsV1
  description: Complete reference of the UpdatePortAssignmentsV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!update-port-assignments
notes:
  - SDK Method used are
    sda.Sda.add_port_assignments_v1,
    sda.Sda.delete_port_assignment_by_id_v1,
    sda.Sda.update_port_assignments_v1,

  - Paths used are
    post /dna/intent/api/v1/sda/portAssignments,
    delete /dna/intent/api/v1/sda/portAssignments,
    delete /dna/intent/api/v1/sda/portAssignments/{id},
    put /dna/intent/api/v1/sda/portAssignments,

"""

EXAMPLES = r"""
- name: Create
  cisco.dnac.sda_port_assignments_v1:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: present
    payload:
    - authenticateTemplateName: string
      connectedDeviceType: string
      dataVlanName: string
      fabricId: string
      interfaceDescription: string
      interfaceName: string
      networkDeviceId: string
      securityGroupName: string
      voiceVlanName: string

- name: Update all
  cisco.dnac.sda_port_assignments_v1:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: present
    payload:
    - authenticateTemplateName: string
      connectedDeviceType: string
      dataVlanName: string
      fabricId: string
      id: string
      interfaceDescription: string
      interfaceName: string
      networkDeviceId: string
      scalableGroupName: string
      voiceVlanName: string

- name: Delete all
  cisco.dnac.sda_port_assignments_v1:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: absent
    dataVlanName: string
    fabricId: string
    interfaceName: string
    networkDeviceId: string
    voiceVlanName: string

- name: Delete by id
  cisco.dnac.sda_port_assignments_v1:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: absent
    id: string

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }
"""



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
module: ios_user
author: Trishna Guha (@trishnaguha)
short_description: Module to manage the aggregates of local users.
description:
  - This module provides declarative management of the local usernames configured on
    network devices. It allows playbooks to manage either individual usernames or the
    aggregate of usernames in the current running config. It also supports purging usernames
    from the configuration that are not explicitly defined.
version_added: 1.0.0
notes:
  - Tested against Cisco IOSXE Version 17.3 on CML.
  - This module works with connection C(network_cli).
    See U(https://docs.ansible.com/ansible/latest/network/user_guide/platform_ios.html)
options:
  aggregate:
    description:
      - The set of username objects to be configured on the remote Cisco IOS device.
        The list entries can either be the username or a hash of username and properties.
        This argument is mutually exclusive with the C(name) argument.
    aliases:
      - users
      - collection
    type: list
    elements: dict
    suboptions:
      name:
        description:
          - The username to be configured on the Cisco IOS device. This argument accepts
            a string value and is mutually exclusive with the C(aggregate) argument. Please
            note that this option is not same as C(provider username).
        type: str
        required: true
      configured_password:
        description:
          - The password to be configured on the Cisco IOS device. The password needs to
            be provided in clear and it will be encrypted on the device. Please note that
            this option is not same as C(provider password).
        type: str
      update_password:
        description:
          - Since passwords are encrypted in the device running config, this argument will
            instruct the module when to change the password.  When set to C(always), the
            password will always be updated in the device and when set to C(on_create) the
            password will be updated only if the username is created.
        choices:
          - on_create
          - always
        type: str
      password_type:
        description:
          - This argument determines whether a 'password' or 'secret' will be configured.
        choices:
          - secret
          - password
        type: str
      hashed_password:
        description:
          - This option allows configuring hashed passwords on Cisco IOS devices.
        type: dict
        suboptions:
          type:
            description:
              - Specifies the type of hash (e.g., 5 for MD5, 8 for PBKDF2, etc.)
              - For this to work, the device needs to support the desired hash type
            type: int
            required: true
          value:
            description:
              - The actual hashed password to be configured on the device
            required: true
            type: str
      privilege:
        description:
          - The C(privilege) argument configures the privilege level of the user when logged
            into the system. This argument accepts integer values in the range of 1 to 15.
        type: int
      view:
        description:
          - Configures the view for the username in the device running configuration. The
            argument accepts a string value defining the view name. This argument does not
            check if the view has been configured on the device.
        aliases:
          - role
        type: str
      sshkey:
        description:
          - Specifies one or more SSH public key(s) to configure for the given username.
          - This argument accepts a valid SSH key value.
        type: list
        elements: str
      nopassword:
        description:
          - Defines the username without assigning a password. This will allow the user
            to login to the system without being authenticated by a password.
        type: bool
      state:
        description:
          - Configures the state of the username definition as it relates to the device
            operational configuration. When set to I(present), the username(s) should be
            configured in the device active configuration and when set to I(absent) the
            username(s) should not be in the device active configuration
        choices:
          - present
          - absent
        type: str
  name:
    description:
      - The username to be configured on the Cisco IOS device. This argument accepts
        a string value and is mutually exclusive with the C(aggregate) argument. Please
        note that this option is not same as C(provider username).
    type: str
  configured_password:
    description:
      - The password to be configured on the Cisco IOS device. The password needs to
        be provided in clear and it will be encrypted on the device. Please note that
        this option is not same as C(provider password).
    type: str
  update_password:
    description:
      - Since passwords are encrypted in the device running config, this argument will
        instruct the module when to change the password.  When set to C(always), the
        password will always be updated in the device and when set to C(on_create) the
        password will be updated only if the username is created.
    default: always
    choices:
      - on_create
      - always
    type: str
  password_type:
    description:
      - This argument determines whether a 'password' or 'secret' will be configured.
    default: secret
    choices:
      - secret
      - password
    type: str
  hashed_password:
    description:
      - This option allows configuring hashed passwords on Cisco IOS devices.
    type: dict
    suboptions:
      type:
        description:
          - Specifies the type of hash (e.g., 5 for MD5, 8 for PBKDF2, etc.)
          - For this to work, the device needs to support the desired hash type
        type: int
        required: true
      value:
        description:
          - The actual hashed password to be configured on the device
        required: true
        type: str
  privilege:
    description:
      - The C(privilege) argument configures the privilege level of the user when logged
        into the system. This argument accepts integer values in the range of 1 to 15.
    type: int
  view:
    description:
      - Configures the view for the username in the device running configuration. The
        argument accepts a string value defining the view name. This argument does not
        check if the view has been configured on the device.
    aliases:
      - role
    type: str
  sshkey:
    description:
      - Specifies one or more SSH public key(s) to configure for the given username.
      - This argument accepts a valid SSH key value.
    type: list
    elements: str
  nopassword:
    description:
      - Defines the username without assigning a password. This will allow the user
        to login to the system without being authenticated by a password.
    type: bool
  purge:
    description:
      - Instructs the module to consider the resource definition absolute. It will remove
        any previously configured usernames on the device with the exception of the
        `admin` user (the current defined set of users).
    type: bool
    default: false
  state:
    description:
      - Configures the state of the username definition as it relates to the device
        operational configuration. When set to I(present), the username(s) should be
        configured in the device active configuration and when set to I(absent) the
        username(s) should not be in the device active configuration
    default: present
    choices:
      - present
      - absent
    type: str
extends_documentation_fragment:
  - cisco.ios.ios
"""

EXAMPLES = """
# Using state: present

# Before state:
# -------------

# router-ios#show running-config | section ^username
# username testuser privilege 15 password 0 password

# Present state create a new user play:
# -------------------------------------

- name: Create a new user
  cisco.ios.ios_user:
    name: ansible
    nopassword: true
    sshkey: "{{ lookup('file', '~/.ssh/id_rsa.pub') }}"
    state: present

# Task Output
# -----------

# commands:
# - ip ssh pubkey-chain
# - username ansible
# - key-hash ssh-rsa 2ABB27BBC33ED53EF7D55037952ABB27 test@fedora
# - exit
# - exit
# - username ansible nopassword

# After state:
# ------------

# router-ios#show running-config | section username
# username testuser privilege 15 password 0 password
# username ansible nopassword
#   username ansible
#    key-hash ssh-rsa 2ABB27BBC33ED53EF7D55037952ABB27 test@fedora

# Using state: present

# Before state:
# -------------

# router-ios#show running-config | section ^username
# username testuser privilege 15 password 0 password

# Present state create a new user with multiple keys play:
# --------------------------------------------------------

- name: Create a new user with multiple keys
  cisco.ios.ios_user:
    name: ansible
    sshkey:
      - "{{ lookup('file', '~/.ssh/id_rsa.pub') }}"
      - "{{ lookup('file', '~/path/to/public_key') }}"
    state: present

# Task Output
# -----------

# commands:
# - ip ssh pubkey-chain
# - username ansible
# - key-hash ssh-rsa 2ABB27BBC33ED53EF7D55037952ABB27 test@fedora
# - key-hash ssh-rsa 1985673DCF7FA9A0F374BB97DC2ABB27 test@fedora
# - exit
# - exit

# After state:
# ------------

# router-ios#show running-config | section username
# username testuser privilege 15 password 0 password
#   username ansible
#    key-hash ssh-rsa 2ABB27BBC33ED53EF7D55037952ABB27 test@fedora
#    key-hash ssh-rsa 1985673DCF7FA9A0F374BB97DC2ABB27 test@fedora

# Using Purge: true

# Before state:
# -------------

# router-ios#show running-config | section ^username
# username admin privilege 15 password 0 password
# username testuser privilege 15 password 0 password
# username ansible nopassword
#   username ansible
#    key-hash ssh-rsa 2ABB27BBC33ED53EF7D55037952ABB27 test@fedora

# Purge all users except admin play:
# ----------------------------------

- name: Remove all users except admin
  cisco.ios.ios_user:
    purge: true

# Task Output
# -----------

# commands:
# - no username testuser
# - no username ansible
# - ip ssh pubkey-chain
# - no username ansible
# - exit

# After state:
# ------------

# router-ios#show running-config | section username
# username admin privilege 15 password 0 password

# Using Purge: true

# Before state:
# -------------

# router-ios#show running-config | section ^username
# username admin privilege 15 password 0 password
# username testuser privilege 15 password 0 password1
# username testuser1 privilege 15 password 0 password2
# username ansible nopassword

# Purge all users except admin and these listed users play:
# ---------------------------------------------------------

- name: Remove all users except admin and these listed users
  cisco.ios.ios_user:
    aggregate:
      - name: testuser
      - name: testuser1
    purge: true

# Task Output
# -----------

# commands:
# - no username ansible

# After state:
# ------------

# router-ios#show running-config | section username
# username admin privilege 15 password 0 password
# username testuser privilege 15 password 0 password1
# username testuser1 privilege 15 password 0 password2

# Using state: present

# Before state:
# -------------

# router-ios#show running-config | section ^username
# username admin privilege 15 password 0 password
# username netop password 0 password1
# username netend password 0 password2

# Present state set multiple users to privilege level 15 play:
# ------------------------------------------------------------

- name: Set multiple users to privilege level 15
  cisco.ios.ios_user:
    aggregate:
      - name: netop
      - name: netend
    privilege: 15
    state: present

# Task Output
# -----------

# commands:
# - username netop privilege 15
# - username netend privilege 15

# After state:
# ------------

# router-ios#show running-config | section username
# username admin privilege 15 password 0 password
# username netop privilege 15 password 0 password1
# username netend privilege 15 password 0 password2

# Using state: present

# Before state:
# -------------

# router-ios#show running-config | section ^username
# username admin privilege 15 password 0 password
# username netop privilege 15 password 0 oldpassword

# Present state Change Password for User netop play:
# --------------------------------------------

- name: Change Password for User netop
  cisco.ios.ios_user:
    name: netop
    configured_password: "newpassword"
    password_type: password
    update_password: always
    state: present

# Task Output
# -----------

# commands:
# - username netop password newpassword

# After state:
# ------------

# router-ios#show running-config | section username
# username admin privilege 15 password 0 password
# username netop privilege 15 password 0 newpassword

# Using state: present

# Before state:
# -------------

# router-ios#show running-config | section ^username
# username admin privilege 15 password 0 password
# username netop privilege 15 password 0 password
# username netend privilege 15 password 0 password

# Present state set user view/role for users play:
# --------------------------------------------

- name: Set user view/role for users
  cisco.ios.ios_user:
    aggregate:
      - name: netop
      - name: netend
    view: network-admin
    state: present

# Task Output
# -----------

# commands:
# - username netop view network-admin
# - username netend view network-admin

# After state:
# ------------

# router-ios#show running-config | section username
# username admin privilege 15 password 0 password
# username netop privilege 15 view network-admin password 0 password
# username netend privilege 15 view network-admin password 0 password

# Using state: present

# Before state:
# -------------

# router-ios#show running-config | section ^username
# username admin privilege 15 password 0 password

# Present state create a new user with hashed password play:
# --------------------------------------------------------------

- name: Create a new user with hashed password
  cisco.ios.ios_user:
    name: ansibletest5
    hashed_password:
      type: 9
      value: "thiswillbereplacedwithhashedpassword"
    state: present

# Task Output
# -----------

# commands:
# - username ansibletest5 secret 9 thiswillbereplacedwithhashedpassword

# After state:
# ------------

# router-ios#show running-config | section username
# username admin privilege 15 password 0 password
# username ansibletest5 secret 9 thiswillbereplacedwithhashedpassword

# Using state: absent

# Before state:
# -------------

# router-ios#show running-config | section ^username
# username admin privilege 15 password 0 password
# username ansibletest1 password 0 password
# username ansibletest2 secret 9 thiswillbereplacedwithhashedpassword
# username ansibletest3 password 5 thistoowillbereplacedwithhashedpassword

# Absent state remove multiple users play:
# ----------------------------------------

- name: Delete users with aggregate
  cisco.ios.ios_user:
    aggregate:
      - name: ansibletest1
      - name: ansibletest2
      - name: ansibletest3
    state: absent

# Task Output
# -----------

# commands:
# - no username ansibletest1
# - no username ansibletest2
# - no username ansibletest3

# After state:
# ------------

# router-ios#show running-config | section username
# username admin privilege 15 password 0 password
"""

RETURN = """
commands:
  description: The list of configuration mode commands to send to the device
  returned: always
  type: list
  sample:
    - username ansible secret password
    - username admin secret admin
"""
import base64
import hashlib
import re

from copy import deepcopy
from functools import partial

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.six import iteritems
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    remove_default_spec,
)

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.ios import (
    get_config,
    load_config,
)


def validate_privilege(value, module):
    if value and not 1 <= value <= 15:
        module.fail_json(msg="privilege must be between 1 and 15, got %s" % value)


def user_del_cmd(username):
    return {
        "command": "no username %s" % username,
        "prompt": "This operation will remove all username related configurations with same name",
        "answer": "y",
        "newline": False,
    }


def add_ssh(command, want, x=None):
    command.append("ip ssh pubkey-chain")
    if x:
        command.append("username %s" % want["name"])
        for item in x:
            command.append("key-hash %s" % item)
        command.append("exit")
    else:
        command.append("no username %s" % want["name"])
    command.append("exit")


def sshkey_fingerprint(sshkey):
    # IOS will accept a MD5 fingerprint of the public key
    # and is easier to configure in a single line
    # we calculate this fingerprint here
    if not sshkey:
        return None
    if " " in sshkey:
        # ssh-rsa AAA...== comment
        keyparts = sshkey.split(" ")
        keyparts[1] = hashlib.md5(base64.b64decode(keyparts[1])).hexdigest().upper()
        return " ".join(keyparts)
    else:
        # just the key, assume rsa type
        return "ssh-rsa %s" % hashlib.md5(base64.b64decode(sshkey)).hexdigest().upper()


def map_obj_to_commands(updates, module):
    commands = list()
    update_password = module.params["update_password"]
    password_type = module.params["password_type"]

    def needs_update(want, have, x):
        return want.get(x) and want.get(x) != have.get(x)

    def add(command, want, x):
        command.append("username %s %s" % (want["name"], x))

    def add_hashed_password(command, want, x, password_type):
        command.append(
            "username %s %s %s %s" % (want["name"], password_type, x.get("type"), x.get("value")),
        )

    for update in updates:
        want, have = update
        if want["state"] == "absent":
            if have["sshkey"]:
                add_ssh(commands, want)
            else:
                commands.append(user_del_cmd(want["name"]))
        if needs_update(want, have, "view"):
            add(commands, want, "view %s" % want["view"])
        if needs_update(want, have, "privilege"):
            add(commands, want, "privilege %s" % want["privilege"])
        if needs_update(want, have, "sshkey"):
            add_ssh(commands, want, want["sshkey"])
        if needs_update(want, have, "configured_password"):
            if update_password == "always" or not have:
                if have and have["password_type"] and password_type != have["password_type"]:
                    module.fail_json(
                        msg="Can not have both a user password and a user secret."
                        + " Please choose one or the other.",
                    )
                add(commands, want, "%s %s" % (password_type, want["configured_password"]))
        if needs_update(want, have, "hashed_password"):
            add_hashed_password(commands, want, want["hashed_password"], password_type)
        if needs_update(want, have, "nopassword"):
            if want["nopassword"]:
                add(commands, want, "nopassword")
            else:
                add(commands, want, user_del_cmd(want["name"]))
    return commands


def parse_view(data):
    match = re.search("view (\\S+)", data, re.M)
    if match:
        return match.group(1)


def parse_sshkey(data, user):
    sshregex = "username %s(\\n\\s+key-hash .+$)+" % user
    sshcfg = re.search(sshregex, data, re.M)
    key_list = []
    if sshcfg:
        match = re.findall("key-hash (\\S+ \\S+(?: .+)?)$", sshcfg.group(), re.M)
        if match:
            key_list = match
    return key_list


def parse_privilege(data):
    match = re.search("privilege (\\S+)", data, re.M)
    if match:
        return int(match.group(1))


def parse_password_type(data):
    type = None
    if data and data.split()[-3] in ["password", "secret"]:
        type = data.split()[-3]
    return type


def map_config_to_obj(module):
    data = get_config(module, flags=["| section username"])
    match = re.findall("(?:^(?:u|\\s{2}u))sername (\\S+)", data, re.M)
    if not match:
        return list()
    instances = list()
    for user in set(match):
        regex = "username %s .+$" % user
        cfg = re.findall(regex, data, re.M)
        cfg = "\n".join(cfg)
        ssh_key_list = parse_sshkey(data, user)
        obj = {
            "name": user,
            "state": "present",
            "nopassword": "nopassword" in cfg,
            "configured_password": None,
            "hashed_password": None,
            "password_type": parse_password_type(cfg),
            "sshkey": ssh_key_list,
            "is_only_ssh_user": False if cfg.strip() and ssh_key_list else True,
            "is_only_normal_user": True if cfg.strip() and ssh_key_list == [] else False,
            "privilege": parse_privilege(cfg),
            "view": parse_view(cfg),
        }
        instances.append(obj)
    return instances


def get_param_value(key, item, module):
    # if key doesn't exist in the item, get it from module.params
    if not item.get(key):
        value = module.params[key]
    # if key does exist, do a type check on it to validate it
    else:
        value_type = module.argument_spec[key].get("type", "str")
        type_checker = module._CHECK_ARGUMENT_TYPES_DISPATCHER[value_type]
        type_checker(item[key])
        value = item[key]
    # validate the param value (if validator func exists)
    validator = globals().get("validate_%s" % key)
    if all((value, validator)):
        validator(value, module)
    return value


def map_params_to_obj(module):
    users = module.params["aggregate"]
    if not users:
        if not module.params["name"] and module.params["purge"]:
            return list()
        elif not module.params["name"]:
            module.fail_json(msg="username is required")
        else:
            aggregate = [{"name": module.params["name"]}]
    else:
        aggregate = list()
        for item in users:
            if not isinstance(item, dict):
                aggregate.append({"name": item})
            elif "name" not in item:
                module.fail_json(msg="name is required")
            else:
                aggregate.append(item)
    objects = list()
    for item in aggregate:
        get_value = partial(get_param_value, item=item, module=module)
        item["configured_password"] = get_value("configured_password")
        item["hashed_password"] = get_value("hashed_password")
        item["nopassword"] = get_value("nopassword")
        item["privilege"] = get_value("privilege")
        item["view"] = get_value("view")
        item["sshkey"] = render_key_list(get_value("sshkey"))
        item["state"] = get_value("state")
        objects.append(item)
    return objects


def render_key_list(ssh_keys):
    key_list = []
    if ssh_keys:
        for item in ssh_keys:
            key_list.append(sshkey_fingerprint(item))
    return key_list


def update_objects(want, have):
    updates = list()
    for entry in want:
        item = next((i for i in have if i["name"] == entry["name"]), None)
        if all((item is None, entry["state"] == "present")):
            updates.append((entry, {}))
        elif item:
            for key, value in iteritems(entry):
                if value and value != item[key]:
                    updates.append((entry, item))
    return updates


def find_set_difference(list1, list2, key):
    want_users = [x[key] for x in list1]
    have_users = [x[key] for x in list2]
    setdifference = set(have_users).difference(want_users)
    result = [item for item in list2 if item[key] in setdifference]
    return result


def main():
    """main entry point for module execution"""
    hashed_password_spec = dict(
        type=dict(type="int", required=True),
        value=dict(no_log=True, required=True),
    )
    element_spec = dict(
        name=dict(),
        configured_password=dict(no_log=True),
        hashed_password=dict(no_log=True, type="dict", options=hashed_password_spec),
        nopassword=dict(type="bool"),
        update_password=dict(default="always", choices=["on_create", "always"]),
        password_type=dict(default="secret", choices=["secret", "password"]),
        privilege=dict(type="int"),
        view=dict(aliases=["role"]),
        sshkey=dict(type="list", elements="str", no_log=False),
        state=dict(default="present", choices=["present", "absent"]),
    )
    aggregate_spec = deepcopy(element_spec)
    aggregate_spec["name"] = dict(required=True)
    # remove default in aggregate spec, to handle common arguments
    remove_default_spec(aggregate_spec)
    argument_spec = dict(
        aggregate=dict(
            type="list",
            elements="dict",
            options=aggregate_spec,
            aliases=["users", "collection"],
        ),
        purge=dict(type="bool", default=False),
    )
    argument_spec.update(element_spec)
    mutually_exclusive = [
        ("name", "aggregate"),
        ("nopassword", "hashed_password", "configured_password"),
    ]
    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=mutually_exclusive,
        supports_check_mode=True,
    )
    warnings = list()
    result = {"changed": False, "warnings": warnings}
    want = map_params_to_obj(module)
    have = map_config_to_obj(module)

    commands = map_obj_to_commands(update_objects(want, have), module)
    if module.params["purge"]:
        setdifference = find_set_difference(want, have, "name")
        for item in setdifference:
            if item["name"] != "admin":
                if item["is_only_ssh_user"]:
                    add_ssh(commands, item)
                if item["is_only_normal_user"]:
                    commands.append(user_del_cmd(item["name"]))
                if item["is_only_normal_user"] is False and item["is_only_ssh_user"] is False:
                    add_ssh(commands, item)
                    commands.append(user_del_cmd(item["name"]))
    result["commands"] = commands
    if commands:
        if not module.check_mode:
            load_config(module, commands)
        result["changed"] = True
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = r"""
---
module: clients_count_info
short_description: Information module for Clients Count Info
description:
- This module represents an alias of the module clients_count_v1_info
version_added: '6.15.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  startTime:
    description:
    - >
      StartTime query parameter. Start time from which API queries the data set related to the resource. It must
      be specified in UNIX epochtime in milliseconds. Value is inclusive. If `startTime` is not provided, API will
      default to current time.
    type: float
  endTime:
    description:
    - >
      EndTime query parameter. End time to which API queries the data set related to the resource. It must be
      specified in UNIX epochtime in milliseconds. Value is inclusive.
    type: float
  type:
    description:
    - >
      Type query parameter. The client device type whether client is connected to network through Wired or
      Wireless medium.
    type: str
  osType:
    description:
    - >
      OsType query parameter. Client device operating system type. This field supports wildcard (`*`) character-
      based search. If the value contains the (`*`) character, please use the /query API for regex search. Ex
      `*iOS*` or `iOS*` or `*iOS` Examples `osType=iOS` (single osType requested) `osType=iOS&osType=Android`
      (multiple osType requested).
    type: str
  osVersion:
    description:
    - >
      OsVersion query parameter. Client device operating system version This field supports wildcard (`*`)
      character-based search. If the value contains the (`*`) character, please use the /query API for regex
      search. Ex `*14.3*` or `14.3*` or `*14.3` Examples `osVersion=14.3` (single osVersion requested)
      `osVersion=14.3&osVersion=10.1` (multiple osVersion requested).
    type: str
  siteHierarchy:
    description:
    - >
      SiteHierarchy query parameter. The full hierarchical breakdown of the site tree starting from Global site
      name and ending with the specific site name. The Root site is named "Global" (Ex.
      "Global/AreaName/BuildingName/FloorName") This field supports wildcard (`*`) character-based search. If the
      value contains the (`*`) character, please use the /query API for regex search. Ex `*BuildingName*` or
      `BuildingName*` or `*BuildingName` Examples `siteHierarchy=Global/AreaName/BuildingName/FloorName` (single
      siteHierarchy requested) `siteHierarchy=Global/AreaName/BuildingName1/FloorName1&siteHierarchy=Global/AreaNa
      me/BuildingName1/FloorName2` (multiple siteHierarchy requested).
    type: str
  siteHierarchyId:
    description:
    - >
      SiteHierarchyId query parameter. The full hierarchy breakdown of the site tree in id form starting from
      Global site UUID and ending with the specific site UUID. (Ex. "globalUuid/areaUuid/buildingUuid/floorUuid")
      This field supports wildcard (`*`) character-based search. Ex `*buildingUuid*` or `buildingUuid*` or
      `*buildingUuid` Examples `siteHierarchyId=globalUuid/areaUuid/buildingUuid/floorUuid` (single
      siteHierarchyId requested) `siteHierarchyId=globalUuid/areaUuid/buildingUuid1/floorUuid1&siteHierarchyId=glo
      balUuid/areaUuid/buildingUuid1/floorUuid2` (multiple siteHierarchyId requested).
    type: str
  siteId:
    description:
    - >
      SiteId query parameter. The site UUID without the top level hierarchy. (Ex."floorUuid") Examples
      `siteId=floorUuid` (single siteId requested) `siteId=floorUuid1&siteId=floorUuid2` (multiple siteId
      requested).
    type: str
  ipv4Address:
    description:
    - >
      Ipv4Address query parameter. IPv4 Address of the network entity either network device or client This field
      supports wildcard (`*`) character-based search. Ex `*1.1*` or `1.1*` or `*1.1` Examples
      `ipv4Address=1.1.1.1` (single ipv4Address requested) `ipv4Address=1.1.1.1&ipv4Address=2.2.2.2` (multiple
      ipv4Address requested).
    type: str
  ipv6Address:
    description:
    - >
      Ipv6Address query parameter. IPv6 Address of the network entity either network device or client This field
      supports wildcard (`*`) character-based search. Ex `*2001 db8*` or `2001 db8*` or `*2001 db8` Examples
      `ipv6Address=2001 db8 0 0 0 0 2 1` (single ipv6Address requested) `ipv6Address=2001 db8 0 0 0 0 2
      1&ipv6Address=2001 db8 85a3 8d3 1319 8a2e 370 7348` (multiple ipv6Address requested).
    type: str
  macAddress:
    description:
    - >
      MacAddress query parameter. The macAddress of the network device or client This field supports wildcard
      (`*`) character-based search. Ex `*AB AB AB*` or `AB AB AB*` or `*AB AB AB` Examples `macAddress=AB AB AB CD
      CD CD` (single macAddress requested) `macAddress=AB AB AB CD CD DC&macAddress=AB AB AB CD CD FE` (multiple
      macAddress requested).
    type: str
  wlcName:
    description:
    - >
      WlcName query parameter. Wireless Controller name that reports the wireless client. This field supports
      wildcard (`*`) character-based search. If the value contains the (`*`) character, please use the /query API
      for regex search. Ex `*wlc-25*` or `wlc-25*` or `*wlc-25` Examples `wlcName=wlc-25` (single wlcName
      requested) `wlcName=wlc-25&wlc-34` (multiple wlcName requested).
    type: str
  connectedNetworkDeviceName:
    description:
    - >
      ConnectedNetworkDeviceName query parameter. Name of the neighbor network device that client is connected to.
      This field supports wildcard (`*`) character-based search. If the value contains the (`*`) character, please
      use the /query API for regex search. Ex `*ap-25*` or `ap-25*` or `*ap-25` Examples
      `connectedNetworkDeviceName=ap-25` (single connectedNetworkDeviceName requested)
      `connectedNetworkDeviceName=ap-25&ap-34` (multiple connectedNetworkDeviceName requested).
    type: str
  ssid:
    description:
    - >
      Ssid query parameter. SSID is the name of wireless network to which client connects to. It is also referred
      to as WLAN ID - Wireless Local Area Network Identifier. This field supports wildcard (`*`) character-based
      search. If the value contains the (`*`) character, please use the /query API for regex search. Ex `*Alpha*`
      or `Alpha*` or `*Alpha` Examples `ssid=Alpha` (single ssid requested) `ssid=Alpha&ssid=Guest` (multiple ssid
      requested).
    type: str
  band:
    description:
    - >
      Band query parameter. WiFi frequency band that client or Access Point operates. Band value is represented in
      Giga Hertz - GHz Examples `band=5GHZ` (single band requested) `band=2.4GHZ&band=6GHZ` (multiple band
      requested).
    type: str
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Clients RetrievesTheTotalCountOfClientsByApplyingBasicFilteringV1
  description: Complete reference of the RetrievesTheTotalCountOfClientsByApplyingBasicFilteringV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!retrieves-the-total-count-of-clients-by-applying-basic-filtering
notes:
  - SDK Method used are
    clients.Clients.retrieves_the_total_count_of_clients_by_applying_basic_filtering_v1,

  - Paths used are
    get /dna/data/api/v1/clients/count,
  - It should be noted that this module is an alias of clients_count_v1_info

"""

EXAMPLES = r"""
- name: Get all Clients Count Info
  cisco.dnac.clients_count_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    startTime: 0
    endTime: 0
    type: string
    osType: string
    osVersion: string
    siteHierarchy: string
    siteHierarchyId: string
    siteId: string
    ipv4Address: string
    ipv6Address: string
    macAddress: string
    wlcName: string
    connectedNetworkDeviceName: string
    ssid: string
    band: string
  register: result

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "count": 0
      },
      "version": "string"
    }
"""



#!/usr/bin/python
# coding: utf-8 -*-

# (c) 2021, Sean Sullivan
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type


ANSIBLE_METADATA = {
    "metadata_version": "1.1",
    "status": ["preview"],
    "supported_by": "community",
}


DOCUMENTATION = """
---
module: workflow_node_wait
author: "Sean Sullivan (@sean-m-sullivan)"
short_description: Wait for a workflow node to finish.
description:
    - Approve an approval node in a workflow job. See
      U(https://www.ansible.com/tower) for an overview.
options:
    workflow_job_id:
      description:
        - ID of the workflow job to monitor for node.
      required: True
      type: int
    name:
      description:
        - Name of the workflow node to wait on.
      required: True
      type: str
    interval:
      description:
        - The interval in sections, to request an update from the controller.
      required: False
      default: 1
      type: float
    timeout:
      description:
        - Maximum time in seconds to wait for a workflow job to reach approval node.
      default: 10
      type: int
extends_documentation_fragment: awx.awx.auth
"""


EXAMPLES = """
- name: Launch a workflow with a timeout of 10 seconds
  workflow_launch:
    workflow_template: "Test Workflow"
    wait: False
  register: workflow

- name: Wait for a workflow node to finish
  workflow_node_wait:
    workflow_job_id: "{{ workflow.id }}"
    name: Approval Data Step
    timeout: 120
"""

RETURN = """

"""


from ..module_utils.controller_api import ControllerAPIModule


def main():
    # Any additional arguments that are not fields of the item can be added here
    argument_spec = dict(
        workflow_job_id=dict(type="int", required=True),
        name=dict(required=True),
        timeout=dict(type="int", default=10),
        interval=dict(type="float", default=1),
    )

    # Create a module for ourselves
    module = ControllerAPIModule(argument_spec=argument_spec)

    # Extract our parameters
    workflow_job_id = module.params.get("workflow_job_id")
    name = module.params.get("name")
    timeout = module.params.get("timeout")
    interval = module.params.get("interval")

    module.wait_on_workflow_node_url(
        url="workflow_jobs/{0}/workflow_nodes/".format(workflow_job_id),
        object_name=name,
        object_type="Workflow Node",
        timeout=timeout,
        interval=interval,
        **{
            "data": {
                "job__name": name,
            }
        }
    )

    # Attempt to look up jobs based on the status
    module.exit_json(**module.json_output)


if __name__ == "__main__":
    main()



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""
native parser

This is the native parser for use with the cli_parse module and action plugin.
The parser functionality used by the network resource modules is leveraged here.

"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
    author: Bradley Thornton (@cidrblock)
    name: native
    short_description: Define configurable options for C(native) sub-plugin of C(cli_parse) module
    description:
    - This plugin documentation provides the configurable options that can be passed
      to the I(ansible.utils.cli_parse) plugins when I(ansible.netcommon.native) is used as a value for
      I(name) option.
    version_added: 1.0.0
"""

EXAMPLES = r"""
- name: "Run command and parse with native"
  ansible.utils.cli_parse:
    command: "show interface"
    parser:
      name: ansible.netcommon.native
    set_fact: POpqMQoJWTiDpEW
  register: nxos_native_command

- name: "Pass text and template_path"
  ansible.utils.cli_parse:
    text: "{{ nxos_native_command['stdout'] }}"
    parser:
      name: ansible.netcommon.native
      template_path: "/home/user/templates/nxos_show_interface.yaml"
  register: nxos_native_text
"""

from ansible.module_utils._text import to_native
from ansible_collections.ansible.utils.plugins.plugin_utils.base.cli_parser import CliParserBase

from ansible_collections.ansible.netcommon.plugins.module_utils.cli_parser.cli_parsertemplate import (
    CliParserTemplate,
)


try:
    import yaml

    try:
        from yaml import CSafeLoader as SafeLoader
    except ImportError:
        from yaml import SafeLoader
    HAS_YAML = True
except ImportError:
    HAS_YAML = False


class CliParser(CliParserBase):
    """The native parser class
    Convert raw text to structured data using the resource module parser
    """

    DEFAULT_TEMPLATE_EXTENSION = "yaml"
    PROVIDE_TEMPLATE_CONTENTS = True

    def parse(self, *_args, **kwargs):
        """Std entry point for a cli_parse parse execution

        :return: Errors or parsed text as structured data
        :rtype: dict

        :example:

        The parse function of a parser should return a dict:
        {"errors": [a list of errors]}
        or
        {"parsed": obj}
        """
        # res = self._check_reqs()
        # if res.get("errors"):
        #     return res

        template_contents = kwargs["template_contents"]
        parser = CliParserTemplate(lines=self._task_args.get("text", "").splitlines())
        try:
            template_obj = yaml.load(template_contents, SafeLoader)
        except Exception as exc:
            return {"errors": [to_native(exc)]}

        try:
            parser.PARSERS = template_obj
            return {"parsed": parser.parse()}
        except Exception as exc:
            msg = "Native parser returned an error while parsing. Error: {err}"
            return {"errors": [msg.format(err=to_native(exc))]}



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = r"""
---
module: tag_member
short_description: Resource module for Tag Member
description:
- This module represents an alias of the module tag_member_v1
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  id:
    description: Id path parameter. Tag ID.
    type: str
  memberId:
    description: MemberId path parameter. TagMember id to be removed from tag.
    type: str
  memberType:
    description: Tag Member's memberType.
    elements: str
    type: list
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Tag AddMembersToTheTagV1
  description: Complete reference of the AddMembersToTheTagV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!add-members-to-the-tag
- name: Cisco DNA Center documentation for Tag RemoveTagMemberV1
  description: Complete reference of the RemoveTagMemberV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!remove-tag-member
notes:
  - SDK Method used are
    tag.Tag.add_members_to_the_tag_v1,
    tag.Tag.remove_tag_member_v1,

  - Paths used are
    post /dna/intent/api/v1/tag/{id}/member,
    delete /dna/intent/api/v1/tag/{id}/member/{memberId},
  - It should be noted that this module is an alias of tag_member_v1

"""

EXAMPLES = r"""
- name: Create
  cisco.dnac.tag_member:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: present
    id: string
    memberType:
    - string
    payload:
      networkinterface:
      - string

- name: Delete by id
  cisco.dnac.tag_member:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: absent
    id: string
    memberId: string

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "version": "string",
      "response": {
        "taskId": "string",
        "url": "string"
      }
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: ec2_vpc_peering
short_description: create, delete, accept, and reject VPC peering connections between two VPCs.
version_added: 1.0.0
version_added_collection: community.aws
description:
  - Read the AWS documentation for VPC Peering Connections
    U(https://docs.aws.amazon.com/AmazonVPC/latest/UserGuide/vpc-peering.html).
options:
  vpc_id:
    description:
      - VPC id of the requesting VPC.
    required: false
    type: str
  peering_id:
    description:
      - Peering connection id.
    required: false
    type: str
  peer_region:
    description:
      - Region of the accepting VPC.
    required: false
    type: str
  peer_vpc_id:
    description:
      - VPC id of the accepting VPC.
    required: false
    type: str
  peer_owner_id:
    description:
      - The AWS account number for cross account peering.
    required: false
    type: str
  state:
    description:
      - Create, delete, accept, reject a peering connection.
    required: false
    default: present
    choices: ['present', 'absent', 'accept', 'reject']
    type: str
  wait:
    description:
      - Wait for peering state changes to complete.
    required: false
    default: false
    type: bool
notes:
  - Support for O(purge_tags) was added in release 2.0.0.
author:
  - Mike Mochan (@mmochan)
  - Alina Buzachis (@alinabuzachis)
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Complete example to create and accept a local peering connection.
- name: Create local account EC2 VPC Peering Connection
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    vpc_id: "vpc-12345678"
    peer_vpc_id: "vpc-87654321"
    state: "present"
    tags:
      Name: "Peering connection for VPC 21 to VPC 22"
      CostCode: "CC1234"
      Project: "phoenix"
  register: vpc_peer

- name: Accept local EC2 VPC Peering request
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    peering_id: "{{ vpc_peer.peering_id }}"
    state: "accept"
  register: action_peer

# Complete example to delete a local peering connection.
- name: Create local account EC2 VPC Peering Connection
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    vpc_id: "vpc-12345678"
    peer_vpc_id: "vpc-87654321"
    state: "present"
    tags:
      Name: "Peering connection for VPC 21 to VPC 22"
      CostCode: "CC1234"
      Project: "phoenix"
  register: vpc_peer

- name: Delete a local EC2 VPC Peering Connection
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    peering_id: "{{ vpc_peer.peering_id }}"
    state: "absent"
  register: vpc_peer

  # Complete example to create and accept a cross account peering connection.
- name: Create cross account EC2 VPC Peering Connection
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    vpc_id: "vpc-12345678"
    peer_vpc_id: "vpc-12345678"
    peer_owner_id: "123456789012"
    state: "present"
    tags:
      Name: "Peering connection for VPC 21 to VPC 22"
      CostCode: "CC1234"
      Project: "phoenix"
  register: vpc_peer

- name: Accept EC2 VPC Peering Connection from remote account
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    peering_id: "{{ vpc_peer.peering_id }}"
    profile: "bot03_profile_for_cross_account"
    state: "accept"
  register: vpc_peer

# Complete example to create and accept an intra-region peering connection.
- name: Create intra-region EC2 VPC Peering Connection
  amazon.aws.ec2_vpc_peering:
    region: "us-east-1"
    vpc_id: "vpc-12345678"
    peer_vpc_id: "vpc-87654321"
    peer_region: "us-west-2"
    state: "present"
    tags:
      Name: "Peering connection for us-east-1 VPC to us-west-2 VPC"
      CostCode: "CC1234"
      Project: "phoenix"
  register: vpc_peer

- name: Accept EC2 VPC Peering Connection from peer region
  amazon.aws.ec2_vpc_peer:
    region: "us-west-2"
    peering_id: "{{ vpc_peer.peering_id }}"
    state: "accept"
  register: vpc_peer

# Complete example to create and reject a local peering connection.
- name: Create local account EC2 VPC Peering Connection
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    vpc_id: "vpc-12345678"
    peer_vpc_id: "vpc-87654321"
    state: "present"
    tags:
      Name: "Peering connection for VPC 21 to VPC 22"
      CostCode: "CC1234"
      Project: "phoenix"
  register: vpc_peer

- name: Reject a local EC2 VPC Peering Connection
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    peering_id: "{{ vpc_peer.peering_id }}"
    state: "reject"

# Complete example to create and accept a cross account peering connection.
- name: Create cross account EC2 VPC Peering Connection
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    vpc_id: "vpc-12345678"
    peer_vpc_id: "vpc-12345678"
    peer_owner_id: "123456789012"
    state: "present"
    tags:
      Name: "Peering connection for VPC 21 to VPC 22"
      CostCode: "CC1234"
      Project: "phoenix"
  register: vpc_peer

- name: Accept a cross account EC2 VPC Peering Connection request
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    peering_id: "{{ vpc_peer.peering_id }}"
    profile: "bot03_profile_for_cross_account"
    state: "accept"
    tags:
      Name: "Peering connection for VPC 21 to VPC 22"
      CostCode: "CC1234"
      Project: "phoenix"

# Complete example to create and reject a cross account peering connection.
- name: Create cross account EC2 VPC Peering Connection
  amazon.aws.ec2_vpc_peering:
    region: "ap-southeast-2"
    vpc_id: "vpc-12345678"
    peer_vpc_id: "vpc-12345678"
    peer_owner_id: "123456789012"
    state: "present"
    tags:
      Name: "Peering connection for VPC 21 to VPC 22"
      CostCode: "CC1234"
      Project: "phoenix"
  register: vpc_peer

- name: Reject a cross account EC2 VPC Peering Connection
  amazon.aws.ec2_vpc_peeriing:
    region: "ap-southeast-2"
    peering_id: "{{ vpc_peer.peering_id }}"
    profile: "bot03_profile_for_cross_account"
    state: "reject"
"""

RETURN = r"""
peering_id:
  description: The id of the VPC peering connection created/deleted.
  returned: always
  type: str
  sample: "pcx-034223d7c0aec3cde"
vpc_peering_connection:
  description: The details of the VPC peering connection.
  returned: success
  type: dict
  contains:
    accepter_vpc_info:
      description: Information about the VPC which accepted the connection.
      returned: success
      type: dict
      contains:
        cidr_block:
          description: The primary CIDR for the VPC.
          returned: when connection is in the accepted state.
          type: str
          sample: "10.10.10.0/23"
        cidr_block_set:
          description: A list of all CIDRs for the VPC.
          returned: when connection is in the accepted state.
          type: list
          elements: dict
          contains:
            cidr_block:
              description: A CIDR block used by the VPC.
              returned: success
              type: str
              sample: "10.10.10.0/23"
        owner_id:
          description: The AWS account that owns the VPC.
          returned: success
          type: str
          sample: "123456789012"
        peering_options:
          description: Additional peering configuration.
          returned: when connection is in the accepted state.
          type: dict
          contains:
            allow_dns_resolution_from_remote_vpc:
              description: Indicates whether a VPC can resolve public DNS hostnames to private IP addresses when queried from instances in a peer VPC.
              returned: success
              type: bool
            allow_egress_from_local_classic_link_to_remote_vpc:
              description: Indicates whether a local ClassicLink connection can communicate with the peer VPC over the VPC peering connection.
              returned: success
              type: bool
            allow_egress_from_local_vpc_to_remote_classic_link:
              description: Indicates whether a local VPC can communicate with a ClassicLink connection in the peer VPC over the VPC peering connection.
              returned: success
              type: bool
        region:
          description: The AWS region that the VPC is in.
          returned: success
          type: str
          sample: "us-east-1"
        vpc_id:
          description: The ID of the VPC
          returned: success
          type: str
          sample: "vpc-0123456789abcdef0"
    requester_vpc_info:
      description: Information about the VPC which requested the connection.
      returned: success
      type: dict
      contains:
        cidr_block:
          description: The primary CIDR for the VPC.
          returned: when connection is not in the deleted state.
          type: str
          sample: "10.10.10.0/23"
        cidr_block_set:
          description: A list of all CIDRs for the VPC.
          returned: when connection is not in the deleted state.
          type: list
          elements: dict
          contains:
            cidr_block:
              description: A CIDR block used by the VPC
              returned: success
              type: str
              sample: "10.10.10.0/23"
        owner_id:
          description: The AWS account that owns the VPC.
          returned: success
          type: str
          sample: "123456789012"
        peering_options:
          description: Additional peering configuration.
          returned: when connection is not in the deleted state.
          type: dict
          contains:
            allow_dns_resolution_from_remote_vpc:
              description: Indicates whether a VPC can resolve public DNS hostnames to private IP addresses when queried from instances in a peer VPC.
              returned: success
              type: bool
            allow_egress_from_local_classic_link_to_remote_vpc:
              description: Indicates whether a local ClassicLink connection can communicate with the peer VPC over the VPC peering connection.
              returned: success
              type: bool
            allow_egress_from_local_vpc_to_remote_classic_link:
              description: Indicates whether a local VPC can communicate with a ClassicLink connection in the peer VPC over the VPC peering connection.
              returned: success
              type: bool
        region:
          description: The AWS region that the VPC is in.
          returned: success
          type: str
          sample: "us-east-1"
        vpc_id:
          description: The ID of the VPC
          returned: success
          type: str
          sample: "vpc-0123456789abcdef0"
    status:
      description: Details of the current status of the connection.
      returned: success
      type: complex
      contains:
        code:
          description: A short code describing the status of the connection.
          returned: success
          type: str
          sample: "active"
        message:
          description: Additional information about the status of the connection.
          returned: success
          type: str
          sample: "Pending Acceptance by 123456789012"
    tags:
      description: Tags applied to the connection.
      returned: success
      type: dict
    expiration_time:
      description: The time that an unaccepted VPC peering connection will expire.
      type: str
      sample: "2024-10-01T12:11:12+00:00"
    vpc_peering_connection_id:
      description: The ID of the VPC peering connection.
      returned: success
      type: str
      sample: "pcx-0123456789abcdef0"
"""

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from typing import Any
from typing import Dict
from typing import NoReturn
from typing import Tuple

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import accept_vpc_peering_connection
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import create_vpc_peering_connection
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import delete_vpc_peering_connection
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_vpc_peering_connections
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import ensure_ec2_tags
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import reject_vpc_peering_connection
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_specifications
from ansible_collections.amazon.aws.plugins.module_utils.transformation import ansible_dict_to_boto3_filter_list


def wait_for_state(client, module: AnsibleAWSModule, state: str, peering_id: str) -> NoReturn:
    waiter = client.get_waiter("vpc_peering_connection_exists")
    filters = {
        "vpc-peering-connection-id": peering_id,
        "status-code": state,
    }
    try:
        waiter.wait(Filters=ansible_dict_to_boto3_filter_list(filters))
    except botocore.exceptions.WaiterError as e:
        module.fail_json_aws(e, "Failed to wait for state change")
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, "Unable to describe Peering Connection while waiting for state to change")


def describe_peering_connections(client, module: AnsibleAWSModule, params) -> Dict[str, Any]:
    peering_connections: Dict = {}

    filters = {
        "requester-vpc-info.vpc-id": params["VpcId"],
        "accepter-vpc-info.vpc-id": params["PeerVpcId"],
    }

    peering_connections = describe_vpc_peering_connections(client, Filters=ansible_dict_to_boto3_filter_list(filters))
    if peering_connections == []:
        # Try again with the VPC/Peer relationship reversed
        filters = {
            "requester-vpc-info.vpc-id": params["PeerVpcId"],
            "accepter-vpc-info.vpc-id": params["VpcId"],
        }
        peering_connections = describe_vpc_peering_connections(
            client, Filters=ansible_dict_to_boto3_filter_list(filters)
        )

    return peering_connections


def is_active(peering_connection: Dict[str, Any]) -> bool:
    return peering_connection["Status"]["Code"] == "active"


def is_rejected(peering_connection: Dict[str, Any]) -> bool:
    return peering_connection["Status"]["Code"] == "rejected"


def is_pending(peering_connection: Dict[str, Any]) -> bool:
    return peering_connection["Status"]["Code"] == "pending-acceptance"


def is_deleted(peering_connection: Dict[str, Any]) -> bool:
    return peering_connection["Status"]["Code"] == "deleted"


def create_peering_connection(client, module: AnsibleAWSModule) -> Tuple[bool, Dict[str, Any]]:
    changed: bool = False
    params: Dict = {}

    params["VpcId"] = module.params.get("vpc_id")
    params["PeerVpcId"] = module.params.get("peer_vpc_id")

    if module.params.get("peer_region"):
        params["PeerRegion"] = module.params["peer_region"]

    if module.params.get("peer_owner_id"):
        params["PeerOwnerId"] = module.params["peer_owner_id"]

    peering_connections = describe_peering_connections(client, module, params)
    for peering_connection in peering_connections:
        changed |= ensure_ec2_tags(
            client,
            module,
            peering_connection["VpcPeeringConnectionId"],
            purge_tags=module.params.get("purge_tags"),
            tags=module.params.get("tags"),
        )

        if is_active(peering_connection):
            return (changed, peering_connection)

        if is_pending(peering_connection):
            return (changed, peering_connection)

    if module.params.get("tags"):
        params["TagSpecifications"] = boto3_tag_specifications(module.params["tags"], types="vpc-peering-connection")

    if module.check_mode:
        return (True, {"VpcPeeringConnectionId": ""})

    peering_connection = create_vpc_peering_connection(client, **params)
    if module.params.get("wait"):
        wait_for_state(client, module, "pending-acceptance", peering_connection["VpcPeeringConnectionId"])
    changed = True
    return (changed, peering_connection)


def delete_peering_connection(client, module: AnsibleAWSModule) -> NoReturn:
    peering_id = module.params.get("peering_id")
    if peering_id:
        peering_connection = get_peering_connection_by_id(client, module, peering_id)
    else:
        params: Dict = {}
        params["VpcId"] = module.params.get("vpc_id")
        params["PeerVpcId"] = module.params.get("peer_vpc_id")
        params["PeerRegion"] = module.params.get("peer_region")

        if module.params.get("peer_owner_id"):
            params["PeerOwnerId"] = module.params["peer_owner_id"]

        peering_connection = describe_peering_connections(client, module, params)[0]

    if not peering_connection:
        module.exit_json(changed=False)
    else:
        peering_id = peering_id or peering_connection["VpcPeeringConnectionId"]

    if is_deleted(peering_connection):
        module.exit_json(msg="Connection in deleted state.", changed=False, peering_id=peering_id)

    if is_rejected(peering_connection):
        module.exit_json(
            msg="Connection has been rejected. State cannot be changed and will be removed automatically by AWS",
            changed=False,
            peering_id=peering_id,
        )

    if not module.check_mode:
        delete_vpc_peering_connection(client, peering_id)
        if module.params.get("wait"):
            wait_for_state(client, module, "deleted", peering_id)

    module.exit_json(changed=True, peering_id=peering_id)


def get_peering_connection_by_id(client, module: AnsibleAWSModule, peering_id: str) -> Dict[str, Any]:
    filters: Dict = {}
    filters["VpcPeeringConnectionIds"] = [peering_id]

    try:
        result = describe_vpc_peering_connections(client, VpcPeeringConnectionIds=[peering_id])
        return result[0]
    except is_boto3_error_code("InvalidVpcPeeringConnectionId.Malformed") as e:
        module.fail_json_aws(e, msg="Malformed connection ID")


def accept_reject_peering_connection(client, module: AnsibleAWSModule, state: str) -> Tuple[bool, Dict[str, Any]]:
    changed: bool = False

    peering_id = module.params.get("peering_id")
    vpc_peering_connection = get_peering_connection_by_id(client, module, peering_id)

    if not (is_active(vpc_peering_connection) or is_rejected(vpc_peering_connection)):
        if not module.check_mode:
            if state == "accept":
                changed |= accept_vpc_peering_connection(client, peering_id)
                target_state = "active"
            else:
                changed |= reject_vpc_peering_connection(client, peering_id)
                target_state = "rejected"

            if module.params.get("wait"):
                wait_for_state(client, module, target_state, peering_id)

        changed = True

    changed |= ensure_ec2_tags(
        client,
        module,
        peering_id,
        purge_tags=module.params.get("purge_tags"),
        tags=module.params.get("tags"),
    )

    # Reload peering conection info to return latest state/params
    vpc_peering_connection = get_peering_connection_by_id(client, module, peering_id)

    return (changed, vpc_peering_connection)


def main():
    argument_spec = dict(
        vpc_id=dict(type="str"),
        peer_vpc_id=dict(type="str"),
        peer_region=dict(type="str"),
        peering_id=dict(type="str"),
        peer_owner_id=dict(type="str"),
        tags=dict(required=False, type="dict", aliases=["resource_tags"]),
        purge_tags=dict(default=True, type="bool"),
        state=dict(default="present", type="str", choices=["present", "absent", "accept", "reject"]),
        wait=dict(default=False, type="bool"),
    )
    required_if = [
        ("state", "present", ["vpc_id", "peer_vpc_id"]),
        ("state", "accept", ["peering_id"]),
        ("state", "reject", ["peering_id"]),
    ]

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True, required_if=required_if)

    state = module.params.get("state")
    peering_id = module.params.get("peering_id")
    vpc_id = module.params.get("vpc_id")
    peer_vpc_id = module.params.get("peer_vpc_id")

    client = module.client("ec2")

    if state == "present":
        (changed, results) = create_peering_connection(client, module)
    elif state == "absent":
        if not peering_id and (not vpc_id or not peer_vpc_id):
            module.fail_json(
                msg="state is absent but one of the following is missing: peering_id or [vpc_id, peer_vpc_id]"
            )

        delete_peering_connection(client, module)
    else:
        (changed, results) = accept_reject_peering_connection(client, module, state)

    formatted_results = camel_dict_to_snake_dict(results)
    # Turn the resource tags from boto3 into an ansible friendly tag dictionary
    formatted_results["tags"] = boto3_tag_list_to_ansible_dict(formatted_results.get("tags", []))

    module.exit_json(
        changed=changed, vpc_peering_connection=formatted_results, peering_id=results["VpcPeeringConnectionId"]
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: license_virtual_account_change_v1
short_description: Resource module for License Virtual Account Change V1
description:
- Manage operation create of the resource License Virtual Account Change V1.
- Transfer devices from one virtual account to another within same smart account.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  device_uuids:
    description: Comma separated device ids.
    elements: str
    type: list
  smart_account_id:
    description: Smart_account_id path parameter. Id of smart account.
    type: str
  virtual_account_name:
    description: Virtual_account_name path parameter. Name of target virtual account.
    type: str
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Licenses ChangeVirtualAccountV1
  description: Complete reference of the ChangeVirtualAccountV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!change-virtual-account
notes:
  - SDK Method used are
    licenses.Licenses.change_virtual_account_v1,

  - Paths used are
    post /dna/intent/api/v1/licenses/smartAccount/{smart_account_id}/virtualAccount/{virtual_account_name}/device/transfer,  # noqa: E501

"""

EXAMPLES = r"""
- name: Create
  cisco.dnac.license_virtual_account_change_v1:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    device_uuids:
    - string
    smart_account_id: string
    virtual_account_name: string

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "taskId": "string",
        "url": "string"
      }
    }
"""



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The ios_lldp_interfaces class
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type


from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.cfg.base import (
    ConfigBase,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import to_list

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.facts.facts import Facts
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.utils.utils import (
    add_command_to_config_list,
    dict_to_set,
    filter_dict_having_none_value,
    normalize_interface,
    remove_command_from_config_list,
    remove_duplicate_interface,
)


class Lldp_Interfaces(ConfigBase):
    """
    The ios_lldp_interfaces class
    """

    gather_subset = ["!all", "!min"]

    gather_network_resources = ["lldp_interfaces"]

    def __init__(self, module):
        super(Lldp_Interfaces, self).__init__(module)

    def get_lldp_interfaces_facts(self, data=None):
        """Get the 'facts' (the current configuration)

        :rtype: A dictionary
        :returns: The current configuration as a dictionary
        """
        facts, _warnings = Facts(self._module).get_facts(
            self.gather_subset,
            self.gather_network_resources,
            data=data,
        )
        lldp_interfaces_facts = facts["ansible_network_resources"].get("lldp_interfaces")
        if not lldp_interfaces_facts:
            return []
        return lldp_interfaces_facts

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        result = {"changed": False}
        commands = list()
        warnings = list()

        if self.state in self.ACTION_STATES:
            existing_lldp_interfaces_facts = self.get_lldp_interfaces_facts()
        else:
            existing_lldp_interfaces_facts = []
        if self.state in self.ACTION_STATES or self.state == "rendered":
            commands.extend(self.set_config(existing_lldp_interfaces_facts))
        if commands and self.state in self.ACTION_STATES:
            if not self._module.check_mode:
                self._connection.edit_config(commands)
            result["changed"] = True
        if self.state in self.ACTION_STATES:
            result["commands"] = commands

        if self.state in self.ACTION_STATES or self.state == "gathered":
            changed_lldp_interfaces_facts = self.get_lldp_interfaces_facts()
        elif self.state == "rendered":
            result["rendered"] = commands
        elif self.state == "parsed":
            running_config = self._module.params["running_config"]
            if not running_config:
                self._module.fail_json(
                    msg="value of running_config parameter must not be empty for state parsed",
                )
            result["parsed"] = self.get_lldp_interfaces_facts(data=running_config)
        else:
            changed_lldp_interfaces_facts = []

        if self.state in self.ACTION_STATES:
            result["before"] = existing_lldp_interfaces_facts
            if result["changed"]:
                result["after"] = changed_lldp_interfaces_facts
        elif self.state == "gathered":
            result["gathered"] = changed_lldp_interfaces_facts

        result["warnings"] = warnings

        return result

    def set_config(self, existing_lldp_interfaces_facts):
        """Collect the configuration from the args passed to the module,
            collect the current configuration (as a dict from facts)

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        config = self._module.params.get("config")
        want = []
        if config:
            for each in config:
                each.update({"name": normalize_interface(each["name"])})
                want.append(each)
        have = existing_lldp_interfaces_facts
        resp = self.set_state(want, have)

        return to_list(resp)

    def set_state(self, want, have):
        """Select the appropriate function based on the state provided

        :param want: the desired configuration as a dictionary
        :param have: the current configuration as a dictionary
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """

        if self.state in ("overridden", "merged", "replaced", "rendered") and not want:
            self._module.fail_json(
                msg="value of config parameter must not be empty for state {0}".format(self.state),
            )

        if self.state == "overridden":
            commands = self._state_overridden(want, have)
        elif self.state == "deleted":
            commands = self._state_deleted(want, have)
        elif self.state in ("merged", "rendered"):
            commands = self._state_merged(want, have)
        elif self.state == "replaced":
            commands = self._state_replaced(want, have)

        return commands

    def _state_replaced(self, want, have):
        """The command generator when state is replaced

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []

        for interface in want:
            for each in have:
                if each["name"] == interface["name"]:
                    break
            else:
                continue
            have_dict = filter_dict_having_none_value(interface, each)
            commands.extend(self._clear_config(dict(), have_dict))
            commands.extend(self._set_config(interface, each))
        # Remove the duplicate interface call
        commands = remove_duplicate_interface(commands)

        return commands

    def _state_overridden(self, want, have):
        """The command generator when state is overridden

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []

        for each in have:
            for interface in want:
                if each["name"] == interface["name"]:
                    break
            else:
                # We didn't find a matching desired state, which means we can
                # pretend we received an empty desired state.
                interface = dict(name=each["name"])
                commands.extend(self._clear_config(interface, each))
                continue
            have_dict = filter_dict_having_none_value(interface, each)
            commands.extend(self._clear_config(dict(), have_dict))
            commands.extend(self._set_config(interface, each))
        # Remove the duplicate interface call
        commands = remove_duplicate_interface(commands)

        return commands

    def _state_merged(self, want, have):
        """The command generator when state is merged

        :rtype: A list
        :returns: the commands necessary to merge the provided into
                  the current configuration
        """
        commands = []

        for interface in want:
            for each in have:
                if interface["name"] == each["name"]:
                    break
            else:
                if self.state == "rendered":
                    commands.extend(self._set_config(interface, dict()))
                continue
            commands.extend(self._set_config(interface, each))

        return commands

    def _state_deleted(self, want, have):
        """The command generator when state is deleted

        :rtype: A list
        :returns: the commands necessary to remove the current configuration
                  of the provided objects
        """
        commands = []

        if want:
            for interface in want:
                for each in have:
                    if each["name"] == interface["name"]:
                        break
                else:
                    continue
                interface = dict(name=interface["name"])
                commands.extend(self._clear_config(interface, each))
        else:
            for each in have:
                commands.extend(self._clear_config(dict(), each))

        return commands

    def _set_config(self, want, have):
        # Set the interface config based on the want and have config
        commands = []

        interface = "interface " + want["name"]
        # Get the diff b/w want and have
        want_dict = dict_to_set(want)
        have_dict = dict_to_set(have)
        diff = want_dict - have_dict

        if diff:
            diff = dict(diff)
            receive = diff.get("receive")
            transmit = diff.get("transmit")
            med_tlv_select = diff.get("med_tlv_select")
            tlv_select = diff.get("tlv_select")
            if receive:
                cmd = "lldp receive"
                add_command_to_config_list(interface, cmd, commands)
            elif receive is False:
                cmd = "no lldp receive"
                add_command_to_config_list(interface, cmd, commands)
            if transmit:
                cmd = "lldp transmit"
                add_command_to_config_list(interface, cmd, commands)
            elif transmit is False:
                cmd = "no lldp transmit"
                add_command_to_config_list(interface, cmd, commands)

            if med_tlv_select:
                med_tlv_select = dict(med_tlv_select)
                if med_tlv_select.get("inventory_management"):
                    add_command_to_config_list(
                        interface,
                        "lldp med-tlv-select inventory-management",
                        commands,
                    )
            if tlv_select:
                tlv_select = dict(tlv_select)
                if tlv_select.get("power_management"):
                    add_command_to_config_list(
                        interface,
                        "lldp tlv-select power-management",
                        commands,
                    )

        return commands

    def _clear_config(self, want, have):
        # Delete the interface config based on the want and have config
        commands = []
        if want.get("name"):
            interface = "interface " + want["name"]
        else:
            interface = "interface " + have["name"]

        if have.get("receive") and have.get("receive") != want.get("receive"):
            cmd = "lldp receive"
            remove_command_from_config_list(interface, cmd, commands)
        if have.get("transmit") and have.get("transmit") != want.get("transmit"):
            cmd = "lldp transmit"
            remove_command_from_config_list(interface, cmd, commands)

        return commands



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_rest_api
short_description: REST API configuration for Cisco Intersight
description:
- Direct REST API configuration for Cisco Intersight.
- All REST API resources and properties must be specified.
- For more information see L(Cisco Intersight,https://intersight.com/apidocs).
extends_documentation_fragment: intersight
options:
  resource_path:
    description:
    - Resource URI being configured related to api_uri.
    type: str
    required: yes
  query_params:
    description:
    - Query parameters for the Intersight API query languange.
    type: dict
  update_method:
    description:
    - The HTTP method used for update operations.
    - Some Intersight resources require POST operations for modifications.
    - json-patch is used for partial updates.
    - json-patch is only supported for patch operations on existing resources and requires the list_body to be a list of dictionaries.
    - See L(The Intersight API Docs, https://intersight.com/apidocs/introduction/methods/) for details on JSON Patch.
    type: str
    choices: [ patch, post, json-patch ]
    default: patch
  api_body:
    description:
    - The paylod for API requests used to modify resources.
    type: dict
  list_body:
    description:
    - The paylod for API requests used to modify resources.
    - Should be used instead of api_body if a list is required in the API payload.
    type: list
    elements: dict
  return_list:
    description:
    - If C(yes), will return a list of API results in the api_response.
    - By default only the 1st element of the API Results list is returned.
    - Can only be used with GET operations.
    type: bool
    default: no
  state:
    description:
    - If C(present), will verify the resource is present and will create if needed.
    - If C(absent), will verify the resource is absent and will delete if needed.
    type: str
    choices: [present, absent]
    default: present
author:
- David Soper (@dsoper2)
- CiscoUcs (@CiscoUcs)
'''

EXAMPLES = r'''
- name: Configure Boot Policy
  intersight_rest_api:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    api_key_uri: "{{ api_key_uri }}"
    validate_certs: "{{ validate_certs }}"
    resource_path: /boot/PrecisionPolicies
    query_params:
      $filter: "Name eq 'vmedia-localdisk'"
    api_body: {
      "Name": "vmedia-localdisk",
      "ConfiguredBootMode": "Legacy",
      "BootDevices": [
        {
          "ObjectType": "boot.VirtualMedia",
          "Enabled": true,
          "Name": "remote-vmedia",
          "Subtype": "cimc-mapped-dvd"
        },
        {
          "ObjectType": "boot.LocalDisk",
          "Enabled": true,
          "Name": "localdisk",
          "Slot": "MRAID",
          "Bootloader": null
        }
      ],
    }
    state: present

- name: Delete Boot Policy
  intersight_rest_api:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    api_key_uri: "{{ api_key_uri }}"
    validate_certs: "{{ validate_certs }}"
    resource_path: /boot/PrecisionPolicies
    query_params:
      $filter: "Name eq 'vmedia-localdisk'"
    state: absent
'''

RETURN = r'''
api_repsonse:
  description: The API response output returned by the specified resource.
  returned: always
  type: dict
  sample:
    "api_response": {
      "BootDevices": [
        {
          "Enabled": true,
          "Name": "remote-vmedia",
          "ObjectType": "boot.VirtualMedia",
          "Subtype": "cimc-mapped-dvd"
        },
        {
          "Bootloader": null,
          "Enabled": true,
          "Name": "boot-lun",
          "ObjectType": "boot.LocalDisk",
          "Slot": "MRAID"
        }
      ],
      "ConfiguredBootMode": "Legacy",
      "Name": "vmedia-localdisk",
      "ObjectType": "boot.PrecisionPolicy",
    }
'''

from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec, compare_values
from ansible.module_utils.basic import AnsibleModule


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        resource_path=dict(type='str', required=True),
        query_params=dict(type='dict'),
        update_method=dict(type='str', choices=['patch', 'post', 'json-patch'], default='patch'),
        api_body=dict(type='dict'),
        list_body=dict(type='list', elements='dict'),
        return_list=dict(type='bool', default=False),
        state=dict(type='str', choices=['absent', 'present'], default='present'),
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
        mutually_exclusive=[
            ['return_list', 'api_body'],
            ['return_list', 'state'],
            ['api_body', 'list_body'],
        ],
    )

    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''

    if module.params['list_body']:
        module.params['api_body'] = module.params['list_body']

    # determine requested operation (config, delete, or neither (get resource only))
    if module.params['state'] == 'present':
        request_delete = False
        # api_body implies resource configuration through post/patch
        # no api_body implies a get operation
        request_config = bool(module.params['api_body'])
        if request_config and not module.params['query_params']:
            # no query_params will try to create the resource without getting the current state
            get_resource = False
        else:
            get_resource = True
    else:  # state == 'absent'
        # state == 'absent' with no query_params is not permitted to avoid accidental deletion
        if not module.params['query_params']:
            raise ValueError('Please specify query_params when state is absent')
        request_delete = True
        request_config = False
        get_resource = True

    if get_resource:
        # get the current state of the resource
        intersight.get_resource(
            resource_path=module.params['resource_path'],
            query_params=module.params['query_params'],
            return_list=module.params['return_list'],
        )

    moid = None
    resource_values_match = False
    if (request_config or request_delete) and intersight.result['api_response'].get('Moid'):
        # resource exists and moid was returned
        moid = intersight.result['api_response']['Moid']
        if request_config:
            resource_values_match = compare_values(module.params['api_body'], intersight.result['api_response'])
        else:  # request_delete
            intersight.delete_resource(
                moid=moid,
                resource_path=module.params['resource_path'],
            )

    if request_config and not resource_values_match:
        intersight.configure_resource(
            moid=moid,
            resource_path=module.params['resource_path'],
            body=module.params['api_body'],
            query_params=module.params['query_params'],
            update_method=module.params['update_method'],
        )
    if module.params['return_list'] and not isinstance(intersight.result['api_response'], list):
        intersight.result['api_response'] = []

    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (c) 2020 Fred-Sun, (@Fred-Sun)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_privateendpoint_info

version_added: "1.8.0"

short_description: Get private endpoints info

description:
    - Get facts for private endpoints.

options:
    name:
        description:
            - Name of resource group.
        type: str
    resource_group:
        description:
            - Limit results by resource group.
        type: str
    tags:
        description:
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Fred-sun (@Fred-sun)
'''

EXAMPLES = '''
- name: Get facts for one private endpoint
  azure_rm_privateendpoint_info:
    resource_group: myResourceGroup
    name: testprivateendpoint

- name: Get all private endpoint under the resource group
  azure_rm_privateendpoint_info:
    resource_group: myResourceGroup

- name: Get all private endpoint under subscription
  azure_rm_virtualnetwork_info:
    tags:
      - key1:value1
'''

RETURN = '''
state:
    description:
        - List of private endpoint dict with same format as M(azure.azcollection.azure_rm_privateendpoint) module paramter.
    returned: always
    type: complex
    contains:
        id:
            description:
                - Resource ID of the private endpoint.
            sample: /subscriptions/xxx-xxx-xxx/resourceGroups/myResourceGroup/providers/Microsoft.Network/privateEndpoints/ped01
            returned: always
            type: str
        etag:
            description:
                -  A unique read-only string that changes whenever the resource is updated.
            sample: 'W/\"20803842-7d51-46b2-a790-ded8971b4d8a'
            returned: always
            type: str
        network_interfaces:
            description:
                - List ID of the network interfaces.
            returned: always
            type: list
            sample:  ["/subscriptions/xxx-xxx-xxx/resourceGroups/myResourceGroup/providers/Microsoft.Network/networkInterfaces/fredprivateendpoint002.nic"]
        location:
            description:
                - Valid Azure location.
            returned: always
            type: str
            sample: eastus
        tags:
            description:
                - Tags assigned to the resource. Dictionary of string:string pairs.
            returned: always
            type: dict
            sample: { "tag1": "abc" }
        provisioning_state:
            description:
                - Provisioning state of the resource.
            returned: always
            sample: Succeeded
            type: str
        name:
            description:
                - Name of the private endpoint.
            returned: always
            type: str
            sample: ped01
        subnets_id:
            description:
                - Subnets associated with the virtual network.
            returned: always
            type: str
            sample: "/subscriptions/xxx-xxx-xxx/resourceGroups/myResourceGroup/providers/Microsoft.Network/virtualNetworks/fredtestRG-vnet/subnets/default"
        manual_private_link_service_connections:
            description:
                - The resource id of the private endpoint to connect.
            returned: when-used
            type: complex
            contains:
                id:
                    description:
                        - The resource id of the private endpoint to connect.
                    returned: always
                    type: str
                    sample: "/subscriptions/xxx/resourceGroups/testRG/providers/Microsoft.Network/privateEndpoints/ped01/privateLinkServiceConnections/ped01"
                name:
                    description:
                        - The name of the private endpoint connection.
                    returned: always
                    type: str
                    sample: ped_name01
                connection_state:
                    description:
                        - State details of endpoint connection
                    type: complex
                    returned: always
                    contains:
                        description:
                            description:
                                - The reason for approval/rejection of the connection.
                            returned: always
                            type: str
                            sample: "Auto Approved"
                        status:
                            description:
                                - Indicates whether the connection has been Approved/Rejected/Removed by the owner of the service.
                            returned: always
                            type: str
                            sample: Approved
                        actions_required:
                            description:
                                - A message indicating if changes on the service provider require any updates on the consumer.
                            type: str
                            returned: always
                            sample: "This is action_required string"
        private_link_service_connections:
            description:
                - The resource id of the private endpoint to connect.
            returned: when-used
            type: complex
            contains:
                id:
                    description:
                        - The resource id of the private endpoint to connect.
                    returned: always
                    type: str
                    sample: "/subscriptions/xxx/resourceGroups/testRG/providers/Microsoft.Network/privateEndpoints/ped01/privateLinkServiceConnections/ped02"
                name:
                    description:
                        - The name of the private endpoint connection.
                    returned: always
                    type: str
                    sample: ped_name02
                connection_state:
                    description:
                        - State details of endpoint connection
                    type: complex
                    returned: always
                    contains:
                        description:
                            description:
                                - The reason for approval/rejection of the connection.
                            returned: always
                            type: str
                            sample: "Auto Approved"
                        status:
                            description:
                                - Indicates whether the connection has been Approved/Rejected/Removed by the owner of the service.
                            returned: always
                            type: str
                            sample: Approved
                        actions_required:
                            description:
                                - A message indicating if changes on the service provider require any updates on the consumer.
                            type: str
                            returned: always
                            sample: "This is action_required string"
                group_ids:
                    description:
                        - List of group_ids associated with private endpoint
                    returned: always
                    type: list
                    sample: ["postgresqlServer"]
        type:
            description:
                - Resource type.
            returned: always
            type: str
            sample: Microsoft.Network/privateEndpoints
'''

try:
    from azure.core.exceptions import ResourceNotFoundError
except Exception:
    # This is handled in azure_rm_common
    pass

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase


class AzureRMPrivateEndpointInfo(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(type='str'),
            resource_group=dict(type='str'),
            tags=dict(type='list', elements='str')
        )

        self.results = dict(
            changed=False,
            privateendpoints=[]
        )

        self.name = None
        self.resource_group = None
        self.tags = None
        self.results = dict(
            changed=False
        )

        super(AzureRMPrivateEndpointInfo, self).__init__(self.module_arg_spec,
                                                         supports_check_mode=True,
                                                         supports_tags=False,
                                                         facts_module=True)

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()) + ['tags']:
            setattr(self, key, kwargs[key])

        if self.name is not None:
            self.results['privateendpoints'] = self.get_item()
        elif self.resource_group is not None:
            self.results['privateendpoints'] = self.list_resource_group()
        else:
            self.results['privateendpoints'] = self.list_items()

        return self.results

    def get_item(self):
        self.log('Get properties for {0}'.format(self.name))
        item = None
        results = []

        try:
            item = self.network_client.private_endpoints.get(self.resource_group, self.name)
        except ResourceNotFoundError:
            self.log('Could not get info for @(Model.ModuleOperationNameUpper).')
        format_item = self.privateendpoints_to_dict(item)

        if format_item and self.has_tags(format_item['tags'], self.tags):
            results = [format_item]
        return results

    def list_resource_group(self):
        self.log('List items for resource group')
        try:
            response = self.network_client.private_endpoints.list(self.resource_group)
        except ResourceNotFoundError as exc:
            self.fail("Failed to list for resource group {0} - {1}".format(self.resource_group, str(exc)))

        results = []
        for item in response:
            format_item = self.privateendpoints_to_dict(item)
            if self.has_tags(format_item['tags'], self.tags):
                results.append(format_item)
        return results

    def list_items(self):
        self.log('List all for items')
        try:
            response = self.network_client.private_endpoints.list_by_subscription()
        except ResourceNotFoundError as exc:
            self.fail("Failed to list all items - {0}".format(str(exc)))

        results = []
        for item in response:
            format_item = self.privateendpoints_to_dict(item)
            if self.has_tags(format_item['tags'], self.tags):
                results.append(format_item)
        return results

    def privateendpoints_to_dict(self, privateendpoint):
        if privateendpoint is None:
            return None
        results = dict(
            id=privateendpoint.id,
            name=privateendpoint.name,
            location=privateendpoint.location,
            tags=privateendpoint.tags,
            provisioning_state=privateendpoint.provisioning_state,
            type=privateendpoint.type,
            etag=privateendpoint.etag,
            subnet_id=privateendpoint.subnet.id
        )
        if privateendpoint.network_interfaces and len(privateendpoint.network_interfaces) > 0:
            results['network_interfaces'] = []
            for interface in privateendpoint.network_interfaces:
                results['network_interfaces'].append(interface.id)
        if privateendpoint.private_link_service_connections and len(privateendpoint.private_link_service_connections) > 0:
            results['private_link_service_connections'] = []
            for connections in privateendpoint.private_link_service_connections:
                connection = {}
                connection['connection_state'] = {}
                connection['id'] = connections.id
                connection['name'] = connections.name
                connection['type'] = connections.type
                connection['group_ids'] = connections.group_ids
                connection['connection_state']['status'] = connections.private_link_service_connection_state.status
                connection['connection_state']['description'] = connections.private_link_service_connection_state.description
                connection['connection_state']['actions_required'] = connections.private_link_service_connection_state.actions_required
                results['private_link_service_connections'].append(connection)
        if privateendpoint.manual_private_link_service_connections and len(privateendpoint.manual_private_link_service_connections) > 0:
            results['manual_private_link_service_connections'] = []
            for connections in privateendpoint.manual_private_link_service_connections:
                connection = {}
                connection['connection_state'] = {}
                connection['id'] = connections.id
                connection['name'] = connections.name
                connection['type'] = connections.type
                connection['group_ids'] = connections.group_ids
                connection['connection_state']['status'] = connections.manual_private_link_service_connection_state.status
                connection['connection_state']['description'] = connections.manual_private_link_service_connection_state.description
                connection['connection_state']['actions_required'] = connections.manual_private_link_service_connection_state.actions_required
                results['manual_private_link_service_connections'].append(connection)
        return results


def main():
    AzureRMPrivateEndpointInfo()


if __name__ == '__main__':
    main()
